/*
 *	analy.h
 */

/*	Copyright (C) 1999 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef analy_h
#define analy_h

#include "x86dis.h"
#include <stdio.h>
#include "global.h"
#include "common.h"
#include "codeanaly.h"
#include "dataanaly.h"
#include "stddata.h"
#include "area.h"

extern int some_analyser_active;
extern int num_ops_parsed;

class analyser;

class unassembler: public x86dis {
public:
	analyser		*analy;
						unassembler():x86dis(X86_OPSIZE32, X86_ADDRSIZE32) {};
	virtual				~unassembler() {};
			void			init(analyser *A);
			void			load(htfile *f);
	virtual	void			done();

	virtual	void			store(htfile *f);
	virtual	char			*addr_sym_func(CPU_ADDR addr);
};

/***************************************************************************/

typedef enum {dtunknown,	dtbyte, dtword, dtdword,	dtstring,	dtunistring, dtarray} taddrtypetype;

struct taddrtype {
	taddrtypetype	type;
	union {
		dword  length;
		// ...?
	};
};

typedef enum {xrefread, xrefwrite,	xrefoffset, xrefjump, xrefcall} txreftype;

struct txref {
	txref	*next;
	ADDR		addr;
	txreftype	xreftype;
};

struct tcomment {
	tcomment	*next;
	char		*text;
};

struct tlabel;

struct taddr {
	ADDR		addr;
	taddr	*left, *right;
	tlabel	*label;
	txref	*xreflist;
	tcomment	*comments;
	taddrtype	type;
};

// geht besser..
typedef enum {scvalid, scread, scwrite,	screadwrite, sccode, scinitialized} tsectype;

typedef enum {
			brnobranch,						// straight exec. flow
			brjump,
			brreturn,
			brcall,
			brjXX
			} tbranchtype;

typedef enum {	acread, acwrite, acoffset } taccesstype;

struct taccess	{
		bool			indexed;
		bool			write;
		taccesstype 	type;
};

typedef enum { label_func, label_loc, label_data } labeltype;

struct tlabel {
	labeltype	type;
	taddr	*addr;
	char		*name;
	tlabel	*left, *right;
};

/*
 *   internal opcodes are interchanged in this format
 */
#define OPCODE dis_insn

/*
 *	an disabled label will be overwritten as soon as possible,
 *   and never be returned nor saved.
 */
#define DISABLE_LABEL(label) label->addr = NULL;

/*
 *	abstract meens function HAS to be overwritten for using this object
 *	stub meens has not neccessary to be overwritten
 *
 */


class codeanalyser;
class dataanalyser;

class analyser: public object	{
public:
	ADDR			addr;
	tqueue		*addr_queue;
	int			ops_parsed;							// for continueing
	bool			active;
	ADDR			next_explored,	first_explored, last_explored;
	tarea		*explored;
	taddr		*addrs;
	codeanalyser	*code;
	dataanalyser	*data;
	unassembler	*unasm;
	tlabel		*labels;
     int			addrthreshold, labelthreshold;
     int			cur_addr_ops, cur_label_ops;                 // for threshold

						analyser() {};
	virtual				~analyser() {};
			void			init();
			void			load(htfile *f);
	virtual	void			done();

			bool			addlabel(ADDR Addr, char *label, labeltype type, char **oldlabel=0);
			void	 		addcomment(ADDR Addr, int line, char *c);
			void			addxref(ADDR from, ADDR to, txreftype action);
			bool			assignaddrlabel(ADDR Addr, char *prefix, labeltype type);
			bool			assignlabel(ADDR Addr, char *label, labeltype type, char **oldlabel=0);
	virtual	void			beginanalysis(ADDR Addr);
	virtual	ADDR			branchaddr(OPCODE *opcode);				// abstract
	virtual	void			*bufptr(ADDR Addr);						// abstract
			bool	  		continueanalysis();
			void			continueanalysisat(ADDR Addr);
			void			dataaccess(ADDR Addr, taccess	access);
			void			dobranch(tbranchtype branch, OPCODE *opcode, int len);
			void			engage_codeanalyser();
			taddr		*enum_addrs(ADDR Addr);
			tlabel		*enum_labels(char *at);
	virtual	taddrtypetype 	examinedata(ADDR Addr);
	virtual	void			examineopcode(OPCODE *opcode);				// abstract
			taddr		*findaddr(ADDR Addr);
			tlabel		*findlabel(char *label);
			void			freeaddr(taddr *Addr);
			void			freeaddrs(taddr *addrs);
			void			freelabel(tlabel *label);
	virtual	tlabel 		*getaddrlabel(ADDR Addr);
			bool			gotoaddr(ADDR Addr);
	virtual	void 		init_codeanalyser();
	virtual	void			init_dataanalyser();
	virtual	void			init_unasm();                                // abstract
	virtual	tbranchtype 	isbranch(OPCODE *opcode);					// abstract
	virtual	void			log(char *s);							// stub
	virtual	CPU_ADDR 		mapaddr(ADDR Addr);						// stub
			taddr		*newaddr(ADDR Addr);
			tlabel		*newlabel(char *label, taddr *Addr, labeltype type);
	virtual	ADDR			nextvalid(ADDR Addr);                        // abstract
	virtual	void			notifyprogress(ADDR	Addr);				// stub
               void			optimize_addr_tree();
               void			optimize_label_tree();
			void			pushaddr(ADDR Addr);
			bool			savetychangeaddr(ADDR Addr);
			void			setactive(bool mode);
               void			set_addr_tree_optimize_threshold(int threshold);
               void			set_label_tree_optimize_threshold(int threshold);
	virtual	void			store(htfile *f);
	virtual	bool			validaddr(ADDR	Addr, tsectype	action);		// abstract
			bool			validcodeaddr(ADDR Addr);
			bool			validreadaddr(ADDR Addr);
			bool			validwriteaddr(ADDR	Addr);

///  interface only (there's no internal use)

	virtual	bool			getformattedline(ADDR Addr, int line, char *buf, int *length);
	virtual	bool			getformattedline_length(ADDR Addr, int line, int *length);
};

#define INVALID_ADDR 0xffffffff

#define MAX_OPS_PER_CONTINUE 10

#endif
