/*
 *	pe_analy.cc
 */

/*	Copyright (C) 1999 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "global.h"
#include "analy.h"
#include "pe_analy.h"
#include "codeanaly.h"
#include "tools.h"

#include "htapp.h"
#include "htdebug.h"
#include "htpe.h"

char		buffer[1024];

void	pe_analyser::init(ht_pe_shared_data *Pe_shared, htfile *File)
{
	pe_shared = Pe_shared;
	file = File;

	ADDR entry=pe_shared->pe32header.entrypoint_address+pe_shared->pe32header.image_base;

	validarea = new tarea();
	validarea->init();

	analyser::init();

	/////////////

	addcomment(entry, 0, " ");
	addcomment(entry, 0, ";****************************");
	addcomment(entry, 0, ";  program entry point");
	addcomment(entry, 0, ";****************************");
	assignlabel(entry, strdup("entrypoint"), label_loc);

     set_addr_tree_optimize_threshold(100);
     set_label_tree_optimize_threshold(100);
	int export_count=pe_shared->exports.funcs.count();
     int *entropy = random_permutation(export_count);
	for (int i=0; i<export_count; i++) {
		ht_pe_export_function *f=(ht_pe_export_function *)pe_shared->exports.funcs.get(*(entropy+i));
		char label[1024];
		if (f->byname) {
			sprintf(buffer, "; exported function %s, ordinal %04x", f->name, f->ordinal);
			sprintf(label, "exp_%s_%d", f->name, f->ordinal);
		} else {
			sprintf(buffer, "; unnamed exported function, ordinal %04x", f->ordinal);
			sprintf(label, "exp_%d", f->ordinal);
		}
		addcomment(f->function_address, 0, " ");
		addcomment(f->function_address, 0, ";********************************************************");
		addcomment(f->function_address, 0, strdup(buffer));
		addcomment(f->function_address, 0, ";********************************************************");
		pushaddr(f->function_address);
		assignlabel(f->function_address, strdup(label), label_func);
	}
     free(entropy);


	int import_count=pe_shared->imports.funcs.count();
     entropy = random_permutation(import_count);
	for (int i=0; i<import_count; i++) {
		ht_pe_import_function *f=(ht_pe_import_function *)pe_shared->imports.funcs.get(*(entropy+i));
		ht_pe_import_library *d=(ht_pe_import_library *)pe_shared->imports.libs.get(f->dll);
		if (f->byname) {
			sprintf(buffer, "%s:%s", d->name, f->name.name);
		} else {
			sprintf(buffer, "%s:%04x", d->name, f->ordinal);
		}
		assignlabel(f->function_address, strdup(buffer), label_func);
	}
     free(entropy);

	int dimport_count=pe_shared->dimports.funcs.count();
     entropy = random_permutation(dimport_count);
	for (int i=0; i<dimport_count; i++) {
		ht_pe_dimport_function *f=(ht_pe_dimport_function *)pe_shared->dimports.funcs.get(*(entropy+i));
		ht_pe_dimport_library *d=(ht_pe_dimport_library *)pe_shared->dimports.libs.get(f->dll);
		if (f->byname) {
			sprintf(buffer, "%s:%s", d->name, f->name.name);
		} else {
			sprintf(buffer, "%s:%04x", d->name, f->ordinal);
		}
		assignlabel(f->function_address, strdup(buffer), label_func);
	}
     free(entropy);

     set_addr_tree_optimize_threshold(1000);
     set_label_tree_optimize_threshold(1000);
}


void pe_analyser::load(htfile *f)
{
	analyser::load(f);
}

void	pe_analyser::done()
{
	validarea->done();
	delete validarea;
	analyser::done();
}


ADDR pe_analyser::branchaddr(OPCODE *opcode)
{
	ADDR		Addr = INVALID_ADDR;
	char *opcode_str = unasm->str(opcode, 0);
	sscanf(opcode_str, "%*s %lx", &Addr);
	return Addr;
}

void *pe_analyser::bufptr(ADDR Addr)
{
	dword ofs;
	Addr-=pe_shared->pe32header.image_base;
	if (!pe_rva_to_ofs(&pe_shared->sections, Addr, &ofs)) assert(0);
	file->seek(ofs);
	file->read(buffer, 15);
	return (void*)buffer;
}

byte digit(char c)
{
	if ((c>='0') && (c<='9')) return (c-'0');
	if ((c>='a') && (c<='f')) return (c-'a'+10);
	if ((c>='A') && (c<='F')) return (c-'A'+10);
	return 42;
}

void	pe_analyser::examineopcode(OPCODE *opcode)
{
	/* _bad_ code, doesn't even work */
	char *opcode_str = unasm->str(opcode, 0);

	ADDR	offset = 0;
	if (!opcode_str) return;
	while (*opcode_str) {
		if (digit(*opcode_str) <= 9) {
			do {
				offset *= 16;
				offset += digit(*opcode_str++);
			} while ((*opcode_str) && (digit(*opcode_str) <= 15));
			taccess a;
			a.indexed = false;
			a.type = acread;
			if (validaddr(offset, scvalid)) {
				dataaccess(offset, a);
				addxref(offset, addr, xrefoffset);
			}
			offset = 0;
			if (!*opcode_str) return;
		}
		opcode_str++;
	}
}

void pe_analyser::init_codeanalyser()
{
	analyser::init_codeanalyser();
	code->loaddefs("analyser/sign.def");
}

void pe_analyser::init_unasm()
{
	unasm = new unassembler();
	unasm->init(this);
}

tbranchtype pe_analyser::isbranch(OPCODE *opcode)
{
	char *opcode_str = unasm->str(opcode, 0);

	if (*opcode_str=='j') {
		if (*(++opcode_str)=='m') return brjump; else return brjXX;
	} else if (strncmp("call", opcode_str, 4)==0) {
		return brcall;
	} else if (strncmp("ret", opcode_str, 3)==0) {
		return brreturn;
	} else return brnobranch;
}

void pe_analyser::log(char *msg)
{
//	LOG(msg);
}

ADDR pe_analyser::nextvalid(ADDR Addr)
{
	return validarea->findnext(Addr);
}

bool pe_analyser::validaddr(ADDR Addr, tsectype action)
{
	pe_section_headers *sections=&pe_shared->sections;
	int i;
	Addr-=pe_shared->pe32header.image_base;
	pe_rva_to_section(sections, Addr, &i);
	PE_SECTION_HEADER *s=sections->sections+i;
	if (!pe_rva_is_valid(sections, Addr)) return false;
	switch (action) {
		case scvalid:
			return true;
		case scread:
			return s->characteristics & PE_SCN_MEM_READ;
		case scwrite:
			return s->characteristics & PE_SCN_MEM_WRITE;
		case screadwrite:
			return s->characteristics & PE_SCN_MEM_WRITE;
		case sccode:
			return (s->characteristics & PE_SCN_MEM_EXECUTE) || (s->characteristics & PE_SCN_CNT_CODE);
		case scinitialized:
			return s->characteristics & PE_SCN_CNT_INITIALIZED_DATA;
	}
	return false;
}
