/* 
 *	The HTE Editor
 *	asm.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __ASM_H__
#define __ASM_H__

#include "common.h"

#define CPU_X86	1

#define MAX_INSN_SIZE	16

struct CPU_ADDR {
	union {
		dword addr32;
		struct {
			dword lo, hi;
		} addr64;
	};
};

struct asm_code {
	asm_code *next;
	dword size;
	byte data[MAX_INSN_SIZE];
};

typedef void dis_insn;
typedef void asm_insn;

/*
 *	CLASS assembler
 */

class assembler: public object {
protected:
	asm_code *codes;
	asm_code code;
	char error_msg[256];
	bool error;

	void emitbyte(byte b);
	void emitword(word w);
	void emitdword(dword d);
	void free_asm_codes();
	void newcode();
	void pushcode();
public:
	assembler();
	~assembler();
/* new */
STUB		virtual asm_code *encode(asm_insn *asm_insn, CPU_ADDR cur_address);
			   char *get_error_msg();
			   asm_code *shortest(asm_code *codes);
ABSTRACT	virtual int prepare_str(asm_insn *asm_insn, char *s);
			   void set_error_msg(char *format, ...);
};

/*
 *	CLASS disassembler
 */

class disassembler: public object {
public:
	disassembler();
	~disassembler();
/* new */
STUB		virtual char *addr_sym_func(CPU_ADDR addr);
ABSTRACT	virtual dis_insn *decode(byte *code, byte maxlen, CPU_ADDR cur_address);
ABSTRACT	virtual byte getsize(dis_insn *disasm_insn);
STUB		virtual char *str(dis_insn *disasm_insn, int style);
ABSTRACT	virtual char *strf(dis_insn *disasm_insn, int style, char *format);
};

/*****************************************************************************
 *	The strf() format                                                       *
 *****************************************************************************
	String	Action
    --------------------------------------------------
	%x		substitute expression with symbol "x"
	?xy...y	if symbol "x" is defined substitute text between the two "y"s with "x"

	Symbol	Desc
    --------------------------------------------------
	p 		prefix
	n 		name
	1 		first operand
	2 		second operand
	3 		third operand
*/

#define DISASM_STRF_VAR			'%'
#define DISASM_STRF_COND			'?'

#define DISASM_STRF_PREFIX		'p'
#define DISASM_STRF_NAME			'n'
#define DISASM_STRF_FIRST		'1'
#define DISASM_STRF_SECOND		'2'
#define DISASM_STRF_THIRD		'3'

#define DISASM_STRF_DEFAULT_FORMAT	"?p#%p #%n\t%1?2#, %2?3#, %3##"

#endif /* __ASM_H__ */
