/* 
 *	The HTE Editor
 *	htdialog.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "htapp.h"
#include "htdialog.h"
#include "htkeyb.h"
#include "htstring.h"
#include "httools.h"
#include "tools.h"

int execute_dialog(ht_dialog *dialog)
{
	int oldx, oldy;
	screen->getcursor(&oldx, &oldy);
	dialog->setstate(ds_normal);
	dialog->receivefocus();
	dialog->sendmsg(htm_draw, 0);
	screen->show();
	while (dialog->getstate()==ds_normal) {
		if (htkeypressed()) {
			do {
				dialog->sendmsg(htm_keypressed, htgetkey());
			} while (htkeypressed());
			dialog->sendmsg(htm_draw, 0);
			screen->show();
		}

	}
	int state=dialog->getstate();
	screen->setcursor(oldx, oldy);
	if (state!=ds_term_cancel) {
		return state;
	} else {
		return 0;
	}
}

int msgbox(int buttonmask, char *title, char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	bounds b, c;
	app->getbounds(&b);
	b.w=65;
	b.h=7;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	ht_dialog *dialog=new ht_dialog(&b, title);

	c.x=b.x+2;
	c.y=b.y+1;
	c.w=b.w-4;
	c.h=b.h-3;
	ht_statictext *text=new ht_statictext(&c, buf);
	dialog->insert(text);

	int buttons=0;
	if (buttonmask & msgbox_ok) buttons++;
	if (buttonmask & msgbox_cancel) buttons++;
	if (buttonmask & msgbox_yes) buttons++;
	if (buttonmask & msgbox_no) buttons++;

	int pos=0;
	if (buttonmask & msgbox_ok) {
		c.x=b.x+b.w/2-5*(buttons-2*pos);
		c.y=b.y+b.h-2;
		c.w=8;
		c.h=1;
		ht_button *button=new ht_button(&c, "Ok", button_ok);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_yes) {
		c.x=b.x+b.w/2-5*(buttons-2*pos);
		c.y=b.y+b.h-2;
		c.w=8;
		c.h=1;
		ht_button *button=new ht_button(&c, "Yes", button_yes);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_no) {
		c.x=b.x+b.w/2-5*(buttons-2*pos);
		c.y=b.y+b.h-2;
		c.w=8;
		c.h=1;
		ht_button *button=new ht_button(&c, "No", button_no);
		dialog->insert(button);
		pos++;
	}
	if (buttonmask & msgbox_cancel) {
		c.x=b.x+b.w/2-5*(buttons-2*pos);
		c.y=b.y+b.h-2;
		c.w=8;
		c.h=1;
		ht_button *button=new ht_button(&c, "Cancel", button_cancel);
		dialog->insert(button);
		pos++;
	}

	int r=execute_dialog(dialog);
	delete dialog;
	return r;
}

int warning(char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return msgbox(msgbox_ok, "warning", buf);
}

int error(char *format, ...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, format);
	vsprintf(buf, format, arg);
	va_end(arg);
	return msgbox(msgbox_ok, "error", buf);
}

/*
 *	CLASS ht_dialog
 */

ht_dialog::ht_dialog(bounds *b, char *desc)
: ht_window(b, desc, 0)
{
}

ht_dialog::~ht_dialog()
{
}

void ht_dialog::draw()
{
	clear();
	ht_window::draw();
}

int ht_dialog::getstate()
{
	return state;
}

void ht_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_button_pressed) {
		switch (msg->data_int) {
			case button_cancel:
				setstate(ds_term_cancel);
				clearmsg(msg);
				break;
			default:
				setstate(ds_term_ok);
				clearmsg(msg);
				break;
		}
	}
	ht_window::handlemsg(msg);
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Escape:
				setstate(ds_term_cancel);
				clearmsg(msg);
				break;
			case K_Return:
				sendmsg(htm_button_pressed, button_ok);
				clearmsg(msg);
				break;
		}
	}
}

void ht_dialog::setstate(int _state)
{
	state=_state;
}

/*
 *	CLASS ht_inputfield
 */

ht_inputfield::ht_inputfield(bounds *b, char *deftext, int _maxstrlen)
: ht_view(b, hto_selectable | hto_ownbuffer, "some inputfield")
{
	maxstrlen=MAX(b->w*b->h, _maxstrlen);
	text=(char*)malloc(maxstrlen+1);
	*text=0;
	if (deftext) strncpy(text, deftext, maxstrlen);
	text[maxstrlen]=0;
	curchar=text;
	insert=1;
	ofs=0;
	setpalette(pal_grp->get(inputfield_palette_str));
}

ht_inputfield::~ht_inputfield()
{
	if (text) delete text;
}

void ht_inputfield::draw()
{
	int c=focused ? inputfield_color_focused : inputfield_color_unfocused;
	buf->fill(size.x, size.y, size.w, size.h, c, ' ');
	char *t=text+ofs, *u=t;
	int i=0, x;
	if (ofs) print(0, 0, inputfield_color_clip, "<");
	do {
		u=t;
		x=lprint(1, i, c, size.w-2, t);
		t+=x;
		if ((focused) && (curchar>=u) && (curchar<t)) {
			setcursor(curchar-u+1, i);
		}
		i++;
	} while ((*t) && (i<size.h));
	if ((focused) && (t==curchar)) setcursor(curchar-u+1, i-1);
	if (strlen(text)-ofs > (size.w-2)*size.h) {
		print(x+1, i-1, inputfield_color_clip, ">");
	}
}

void ht_inputfield::getdata(void *buf)
{
	((ht_inputfield_data*)buf)->text=text;
}

int ht_inputfield::getdatasize()
{
	return sizeof text;
}

void ht_inputfield::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Up:
				curchar-=size.w-2;
				if (curchar<text) curchar=text;
				if (curchar-text<ofs) ofs=curchar-text;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down: {
				int slen=strlen(text);
				if (slen) {
					curchar+=size.w-2;
					if (curchar-text>slen-1) curchar=text+slen-1;
					if (ofs<curchar-text-(size.w-2)*size.h+1) ofs=curchar-text-(size.w-2)*size.h+1;
				}
				dirtyview();
				clearmsg(msg);
				break;
			}
			case K_Left:
				if (curchar>text) curchar--;
				if (curchar-text<ofs) ofs--;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				if (*curchar) {
					curchar++;
					if (ofs<curchar-text-(size.w-2)*size.h+1) ofs++;
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_BackSpace:
				if (curchar>text) {
					curchar--;
					if (curchar-text<ofs) ofs--;
					memmove(curchar, curchar+1, strlen(curchar));
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_Delete:
				if (*curchar) memmove(curchar, curchar+1, strlen(curchar));
				dirtyview();
				clearmsg(msg);
				break;
			case K_Home:
				curchar=text;
				if (curchar-text<ofs) ofs=curchar-text;
				dirtyview();
				clearmsg(msg);
				break;
			case K_End:
				curchar=text+strlen(text);
				if (curchar-text>=maxstrlen) curchar--;
				if (ofs<curchar-text-(size.w-2)*size.h+1) ofs=curchar-text-(size.w-2)*size.h+1;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Insert:
				insert=!insert;
				dirtyview();
				clearmsg(msg);
				break;
			default:
				if ((msg->data_int>=32) && (msg->data_int<=0xff)) {
					setchar(msg->data_int, insert);
					dirtyview();
					clearmsg(msg);
				}
		}
	}
	ht_view::handlemsg(msg);
}

void ht_inputfield::setchar(char a, int insert)
{
	if (curchar-text<maxstrlen) {
		if (insert) {
			int l=strlen(curchar);
			if (l+curchar-text<maxstrlen) {
				memmove(curchar+1, curchar, l+1);
				if (!*curchar) *(curchar+1)=0;
				*curchar=a;
				if (curchar-text<maxstrlen-1) curchar++;
			}
		} else {
			if (!*curchar) *(curchar+1)=0;
			*curchar=a;
			if (curchar-text<maxstrlen-1) curchar++;
		}
		if (ofs<curchar-text-(size.w-2)*size.h+1) ofs++;
	}
}

/*
 *	CLASS ht_inputfield_button
 */

ht_inputfield_button::ht_inputfield_button(bounds *b, char *deftext, int _maxstrlen, int _value)
: ht_inputfield(b, deftext, _maxstrlen)
{
	value=_value;
	pressed=0;
}

ht_inputfield_button::~ht_inputfield_button()
{
}

void ht_inputfield_button::getdata(void *buf)
{
	((ht_inputfield_button_data*)buf)->text=text;
	((ht_inputfield_button_data*)buf)->pressed=pressed;
}

int ht_inputfield_button::getdatasize()
{
	return sizeof text+sizeof pressed;
}

void ht_inputfield_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Return:
			case K_Space:
				group->sendmsg(htm_button_pressed, value);
				pressed=1;
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_inputfield::handlemsg(msg);
}

/*
 *	CLASS ht_button
 */

ht_button::ht_button(bounds *b, char *_text, int _value)
: ht_view(b, hto_selectable | hto_ownbuffer, "some button")
{
	value=_value;
	text=_text;
	pressed=0;
	setpalette(pal_grp->get(button_palette_str));
}

ht_button::~ht_button()
{
}

void ht_button::draw()
{
	int c=focused ? button_color_focused : button_color_unfocused;
	buf->fill(size.x, size.y, size.w, size.h, c, ' ');
	print((size.w-strlen(text))/2, size.h/2, c, text);
}

void ht_button::getdata(void *buf)
{
	((ht_button_data*)buf)->pressed=pressed;
}

int ht_button::getdatasize()
{
	return sizeof value;
}

void ht_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Return:
			case K_Space:
				group->sendmsg(htm_button_pressed, value);
				pressed=1;
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_listbox
 */

ht_listbox::ht_listbox(bounds *b)
: ht_group(b, hto_selectable, "some listbox")
{
	bounds c;

	c=*b;
	c.x+=c.w-1;
	c.w=1;
	scrollbar=new ht_scrollbar(&c);
	scrollbar->disable();
	insert(scrollbar);

	c=*b;
	c.w--;
	body=new ht_listbox_body(&c, scrollbar);
	insert(body);
}

ht_listbox::~ht_listbox()
{
}

void ht_listbox::getdata(void *buf)
{
	body->getdata(buf);
}

int ht_listbox::getdatasize()
{
	return body->getdatasize();
}

void ht_listbox::insertstring(char *str)
{
	return body->insertstring(str);
}

int ht_listbox::seek(int index)
{
	return body->seek(index);
}

void ht_listbox::setpalette(palette *palette)
{
	body->setpalette(palette);
}

/*
 *	CLASS ht_listbox_popup
 */

ht_listbox_body::ht_listbox_body(bounds *b, ht_scrollbar *_scrollbar)
: ht_view(b, hto_selectable | hto_ownbuffer, "some listbox_body")
{
	scrollbar=_scrollbar;
	list=new ht_string_collection();
	pos=0;
	cursor=0;
	setpalette(pal_grp->get(listbox_palette_str));
	vis_h_size=0;
	bounds c=*b;
	c.x+=c.w-1;
	c.w=1;
}

ht_listbox_body::~ht_listbox_body()
{
	delete list;
}

void ht_listbox_body::draw()
{
	int w=size.w;
	int l=list->count();
	int d=l-pos;
	if (d>size.h) d=size.h;
	if (l>size.h) {
		w--;
	}
	buf->fill(size.x, size.y, size.w, size.h, listbox_color_unselected, ' ');
	for (int i=0; i<d; i++) {
		int c=(i==cursor) ? (focused ? listbox_color_focused_selected : listbox_color_unfocused_selected) : listbox_color_unselected;
		if (i==cursor) {
			buf->fill(size.x+0, size.y+i, w, 1, c, ' ');
		}
		lprint(0, i, c, w, list->get(i+pos));
	}
}

void ht_listbox_body::getdata(void *buf)
{
	((ht_listbox_data*)buf)->list=list;
	((ht_listbox_data*)buf)->index=pos+cursor;
}

int ht_listbox_body::getdatasize()
{
	return 4+4;
}

void ht_listbox_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Control_PageUp:
			case K_Home:
				pos=0;
				cursor=0;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Control_PageDown:
			case K_End: {
				pos=0;
				int c=list->count();
				cursor=c-1;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
				}
				dirtyview();
				clearmsg(msg);
				break;
			}
			case K_PageUp:
				cursor-=vis_h_size;
				if (cursor<0) {
					pos+=cursor;
					cursor=0;
					if (pos<0) pos=0;
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_PageDown: {
				int c=list->count();
				cursor+=vis_h_size;
				if (cursor>=vis_h_size) {
					pos+=cursor-vis_h_size+1;
					cursor=vis_h_size-1;
					if (pos+cursor>=c) pos=c-cursor-1;
				}
				dirtyview();
				clearmsg(msg);
				break;
			}
			case K_Up:
				if (cursor) {
					cursor--;
				} else {
					if (pos) pos--;
				}
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (cursor<vis_h_size-1) {
					if (cursor<list->count()-1) cursor++;
				} else {
					if (pos+vis_h_size<list->count()) pos++;
				}
				dirtyview();
				clearmsg(msg);
				break;
		}
		int c=list->count();
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
	}
	ht_view::handlemsg(msg);
}

void ht_listbox_body::insertstring(char *str)
{
	list->add(str);
	int c=list->count();
	vis_h_size=MIN(size.h, c);
	if ((c>size.h) && (!(scrollbar->enabled))) scrollbar->enable();
}

int ht_listbox_body::seek(int index)
{
	int c=list->count();
	if (index<c) {
		cursor=index;
		pos=0;
		if (cursor>vis_h_size-1) {
			pos=cursor+1-vis_h_size;
			cursor=vis_h_size-1;
		}
		if (c>1) scrollbar->setpos((pos+cursor)*100/(c-1));
		return 1;
	}
	return 0;
}

/*
 *	CLASS ht_listbox_button
 */

ht_listbox_button::ht_listbox_button(bounds *b, int _value)
: ht_listbox(b)
{
	value=_value;
}

ht_listbox_button::~ht_listbox_button()
{
}

void ht_listbox_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Return:
			case K_Space:
				group->sendmsg(htm_button_pressed, value);
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_listbox::handlemsg(msg);
}

/*
 *	CLASS ht_listpopup
 */

ht_listpopup::ht_listpopup(bounds *b)
: ht_group(b, hto_selectable, "some listpopup")
{
	bounds c;

	c=*b;
	c.x+=c.w-1;
	c.w=1;
	scrollbar=new ht_scrollbar(&c);
	scrollbar->disable();
	insert(scrollbar);

	c=*b;
	c.w--;
	c.h=1;
	body=new ht_listpopup_body(&c, scrollbar);
	insert(body);
}

ht_listpopup::~ht_listpopup()
{
}

void ht_listpopup::getdata(void *buf)
{
	body->getdata(buf);
}

int ht_listpopup::getdatasize()
{
	return body->getdatasize();
}

void ht_listpopup::insertstring(char *str)
{
	return body->insertstring(str);
}

void ht_listpopup::setpalette(palette *palette)
{
	body->setpalette(palette);
}


/*
 *	CLASS ht_listpopup_body
 */

#define LISTPOPUP_HEIGHT 5

ht_listpopup_body::ht_listpopup_body(bounds *b, ht_scrollbar *scrollbar)
: ht_listbox_body(b, scrollbar)
{
	disable_buffering();
	expanded=0;
}

ht_listpopup_body::~ht_listpopup_body()
{
}

void ht_listpopup_body::collapse()
{
	if (expanded) {
		expanded=0;
		if (scrollbar->enabled) {
			scrollbar->disable();
		}
		dirtyview();
	}
}

void ht_listpopup_body::draw()
{
	int c=listbox_color_unfocused_selected;
	if (focused) {
		c=listbox_color_focused_selected;
	}
	buf->fill(size.x, size.y, size.w, 1, c, ' ');
	lprint(0, 0, c, size.w-1, list->get(pos+cursor));
	printchar(size.w-1, 0, listbox_color_unselected, CHAR_ARROW_DOWN);
	if (expanded) {
		buf->fill(size.x, size.y+1, size.w, 1, listbox_color_unselected, CHAR_LINEH);
		buf->fill(size.x, size.y+2, size.w, vis_h_size, listbox_color_unselected, ' ');
		for (int i=0; i<vis_h_size; i++) {
			int c=listbox_color_unselected;
			if (i==cursor) {
				c=listbox_color_focused_selected;
				buf->fill(size.x, size.y+i+2, size.w, 1, c, ' ');
			}
			lprint(0, i+2, c, size.w, list->get(pos+i));
		}
	}
}


void ht_listpopup_body::expand()
{
	if (!expanded) {
		bounds b;
		getbounds(&b);
		b.h=MIN(LISTPOPUP_HEIGHT, list->count())+2;
		setbounds(&b);
		expanded=1;
		vis_h_size=MIN(size.h-2, list->count());
		dirtyview();
		if ((!(scrollbar->enabled)) && (list->count()>vis_h_size)) {
			bounds b;
			scrollbar->getbounds(&b);
			b.y=size.y+2;
			b.h=vis_h_size;
			scrollbar->setbounds(&b);
			scrollbar->enable();
		}
	}
}

void ht_listpopup_body::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
			case K_Right:
			case K_Up:
			case K_Down:
			case K_PageUp:
			case K_PageDown:
			case K_Control_PageUp:
			case K_Control_PageDown:
			case K_Home:
			case K_End:
				expand();
				break;
			case K_Return:
				if (expanded) {
					collapse();
					dirtyview();
					clearmsg(msg);
				}
				break;
		}
	}
	ht_listbox_body::handlemsg(msg);
}

void ht_listpopup_body::insertstring(char *str)
{
	list->add(str);
	int c=list->count();
	vis_h_size=MIN(size.h, c);
}

void ht_listpopup_body::releasefocus()
{
	collapse();
	ht_listbox_body::releasefocus();
}

/*
 *	CLASS ht_statictext
 */

ht_statictext::ht_statictext(bounds *b, char *_text)
: ht_view(b, hto_ownbuffer, "some statictext")
{
	text=strdup(_text);
	setpalette(pal_grp->get(window_palette_str));
}

ht_statictext::~ht_statictext()
{
	if (text) delete text;
}

void ht_statictext::draw()
{
	buf->fill(size.x, size.y, size.w, size.h, window_color_normal, ' ');
	char *t=text, *u=t;
	int i=0;
	do {
		u=t;
		t+=lprint(0, i, window_color_normal, size.w, t);
		i++;
	} while ((*t) && (i<size.h));
}

void ht_statictext::getdata(void *buf)
{
	((ht_statictext_data*)buf)->text=text;
}

int ht_statictext::getdatasize()
{
	return sizeof text;
}

/*
 *	CLASS ht_listbox_ptr
 */

ht_listbox_ptr::ht_listbox_ptr(ht_listbox *_listbox)
{
	listbox=_listbox;
}

ht_listbox_ptr::~ht_listbox_ptr()
{
}

/*
 *	CLASS ht_color_block
 */

ht_color_block::ht_color_block(bounds *b, int light_colors, int selected)
: ht_view(b, hto_ownbuffer | hto_selectable, "some color_block")
{
	color=selected;
	if (light_colors) colors=16; else colors=8;
}

ht_color_block::~ht_color_block()
{
}

void ht_color_block::draw()
{
	clear();
	for (int i=0; i<colors; i++) {
		if (i==color) {
			int fgc=C_WHITE;
			if (focused) fgc=C_LIGHT(C_WHITE);
			printchar((i%4)*2, i/4, pal_grp->get_color(fgc, C_BLACK), '>');
			printchar((i%4)*2+1, i/4, pal_grp->get_color(i, 0), CHAR_FILLED_M);
			printchar((i%4)*2+2, i/4, pal_grp->get_color(fgc, C_BLACK), '<');
		} else {
			printchar((i%4)*2+1, i/4, pal_grp->get_color(i, 0), CHAR_FILLED_M);
		}
	}
}

void ht_color_block::getdata(void *buf)
{
	((ht_color_block_data*)buf)->color=color;
}

int ht_color_block::getdatasize()
{
	return sizeof color;
}

void ht_color_block::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Left:
				if (color%4-1>=0) color--;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Right:
				if (color%4+1<=3) color++;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Up:
				if (color-4>=0) color-=4;
				dirtyview();
				clearmsg(msg);
				break;
			case K_Down:
				if (color+4<colors) color+=4;
				dirtyview();
				clearmsg(msg);
				break;
		}
	}
	ht_view::handlemsg(msg);
}

void ht_color_block::set_color(int i)
{
	color=i;
	dirtyview();
}

/*
 *	CLASS ht_color_manager
 */

ht_color_manager::ht_color_manager(bounds *b, palette_group *_pal_grp)
:ht_dialog(b, "colors")
{
	ipalette *i;

	orig_pal_grp=_pal_grp;
	pal_grp=orig_pal_grp->copy();

	bounds c;
/* group text */
	c.x=b->x+2;
	c.y=b->y+1;
	c.w=20;
	c.h=2;
	insert(new ht_statictext(&c, "group"));
/* group */
	c.x=b->x+2;
	c.y=b->y+2;
	c.w=15;
	c.h=7;
	group=new ht_listbox(&c);
	i=pal_grp->enum_first();
	while (i) {
		group->insertstring(i->name);
		i=pal_grp->enum_next();
	}
	insert(group);

/* items text */
	c.x=b->x+18;
	c.y=b->y+1;
	c.w=10;
	c.h=1;
	insert(new ht_statictext(&c, "item"));
/* items */
	c.x=b->x+18;
	c.y=b->y+2;
	c.w=20;
	c.h=7;
	items=new ht_xgroup(&c, hto_selectable, "items");
	group_to_items=new ht_collection();
	i=pal_grp->enum_first();
	while (i) {
		palette *p=i->pal;
		ht_listbox *itemlist=new ht_listbox(&c);
		group_to_items->add(new ht_listbox_ptr(itemlist));
		for (int k=0; k<i->entries; k++) {
			itemlist->insertstring(p->name);
			p++;
		}
		items->insert(itemlist);
		i=pal_grp->enum_next();
	}
	insert(items);
/* foreground text */
	c.x=b->x+40;
	c.y=b->y+2;
	c.w=8;
	c.h=1;
	insert(new ht_statictext(&c, "fg color"));
/* foreground colors */
	c.x=b->x+40;
	c.y=b->y+3;
	c.w=9;
	c.h=4;
	fg_colors=new ht_color_block(&c, 1, 0);
	insert(fg_colors);
/* background text */
	c.x=b->x+40;
	c.y=b->y+7;
	c.w=8;
	c.h=1;
	insert(new ht_statictext(&c, "bg color"));
/* background colors */
	c.x=b->x+40;
	c.y=b->y+8;
	c.w=9;
	c.h=2;
	bg_colors=new ht_color_block(&c, 0, 0);
	insert(bg_colors);
/* select the current item */
	sendmsg(htm_keypressed, K_Up);	/* FIXME: this a hack ! */
}

ht_color_manager::~ht_color_manager()
{
	delete group_to_items;
	delete pal_grp;
	app->sendmsg(htm_dirtyview, 0);
}

void ht_color_manager::getdata(void *buf)
{
	((ht_color_manager_data*)buf)->pal_grp=pal_grp;
}

int ht_color_manager::getdatasize()
{
	return sizeof pal_grp;
}

void ht_color_manager::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {

		ht_dialog::handlemsg(msg);

		if (group->isviewdirty() || items->getcurrentchild()->isviewdirty()) {
			ht_view *cur=current;
			ht_listbox_data data;
			group->getdata(&data);
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(data.index);
			char *s=data.list->get(data.index);
			palette *q=pal_grp->get(s);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				char color;
				if (pal_grp->get_entry(q, data.index, &name, &color)) {
					fg_colors->set_color(color&0xf);
					bg_colors->set_color((color>>4)&0x7);
				}
			}
			if (group->isviewdirty()) {
				focus(p->listbox);
				focus(cur);
			}
		} else if (fg_colors->isviewdirty() || bg_colors->isviewdirty()) {
			ht_listbox_data data;
			group->getdata(&data);
			ht_listbox_ptr *p=(ht_listbox_ptr*)group_to_items->get(data.index);
			char *s=data.list->get(data.index);
			palette *q=pal_grp->get(s);
			if (q) {
				p->listbox->getdata(&data);
				char *name;
				char color;
				if (pal_grp->get_entry(q, data.index, &name, &color)) {
					ht_color_block_data d;
					fg_colors->getdata(&d);
					color=d.color;
					bg_colors->getdata(&d);
					color|=d.color<<4;
					pal_grp->set_entry(q, data.index, name, color);
				}
			}
		}
	} else {
		ht_dialog::handlemsg(msg);
	}
}

/*
 *	INIT
 */

int init_dialog()
{
	palette *inputfield_palette=pal_grp->create(inputfield_palette_str, inputfield_palette_colors);
	pal_grp->set_entry(inputfield_palette, inputfield_coloridx_focused, "focused", COLOR(C_LIGHT(C_WHITE), C_BLUE));
	pal_grp->set_entry(inputfield_palette, inputfield_coloridx_unfocused, "unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE));
	pal_grp->set_entry(inputfield_palette, inputfield_coloridx_clip, "clipping char (<, >)", COLOR(C_LIGHT(C_GREEN), C_BLUE));

	palette *button_palette=pal_grp->create(button_palette_str, button_palette_colors);
	pal_grp->set_entry(button_palette, button_coloridx_focused, "focused", COLOR(C_LIGHT(C_WHITE), C_GREEN));
	pal_grp->set_entry(button_palette, button_coloridx_unfocused, "unfocused", COLOR(C_BLACK, C_GREEN));

	palette *listbox_cyan_palette=pal_grp->create(listbox_cyan_palette_str, listbox_palette_colors);
	pal_grp->set_entry(listbox_cyan_palette, listbox_coloridx_unselected, "unselected", COLOR(C_BLACK, C_CYAN));
	pal_grp->set_entry(listbox_cyan_palette, listbox_coloridx_focused_selected, "selected & focused", COLOR(C_LIGHT(C_WHITE), C_GREEN));
	pal_grp->set_entry(listbox_cyan_palette, listbox_coloridx_unfocused_selected, "selected & unfocused", COLOR(C_LIGHT(C_WHITE), C_CYAN));

	palette *listbox_black_palette=pal_grp->create(listbox_black_palette_str, listbox_palette_colors);
	pal_grp->set_entry(listbox_black_palette, listbox_coloridx_unselected, "unselected", COLOR(C_WHITE, C_BLACK));
	pal_grp->set_entry(listbox_black_palette, listbox_coloridx_focused_selected, "selected & focused", COLOR(C_LIGHT(C_WHITE), C_BLUE));
	pal_grp->set_entry(listbox_black_palette, listbox_coloridx_unfocused_selected, "selected & unfocused", COLOR(C_LIGHT(C_WHITE), C_BLACK));

	return 1;
}

/*
 *	DONE
 */

int done_dialog()
{
	return 1;
}

