/* 
 *	The HTE Editor
 *	htdialog.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTDIALOG_H__
#define __HTDIALOG_H__

#include "htdebug.h"
#include "htobj.h"

/*
 *	Palette entries
 */

#define inputfield_palette_str 		"inputfield"
#define inputfield_palette_colors       3
#define inputfield_coloridx_focused	0
#define inputfield_coloridx_unfocused	1
#define inputfield_coloridx_clip		2
#define inputfield_color_focused		pal[inputfield_coloridx_focused].value
#define inputfield_color_unfocused		pal[inputfield_coloridx_unfocused].value
#define inputfield_color_clip			pal[inputfield_coloridx_clip].value

#define button_palette_str			"button"
#define button_palette_colors			2
#define button_coloridx_focused		0
#define button_coloridx_unfocused		1
#define button_color_focused			pal[button_coloridx_focused].value
#define button_color_unfocused		pal[button_coloridx_unfocused].value

#define listbox_black_palette_str			"listbox (cyan)"
#define listbox_cyan_palette_str			"listbox (black)"

#define listbox_palette_str				listbox_cyan_palette_str
#define listbox_palette_colors          	3
#define listbox_coloridx_unselected		0
#define listbox_coloridx_focused_selected	1
#define listbox_coloridx_unfocused_selected	2
#define listbox_color_unselected			pal[listbox_coloridx_unselected].value
#define listbox_color_focused_selected		pal[listbox_coloridx_focused_selected].value
#define listbox_color_unfocused_selected	pal[listbox_coloridx_unfocused_selected].value

/*
 *	CLASS ht_dialog
 */

#define ds_normal			0
#define ds_term_ok			1
#define ds_term_cancel		2

#define button_cancel		0
#define button_ok			1
#define button_yes            button_ok
#define button_no			2

class ht_dialog: public ht_window {
private:
	int state;
public:
	ht_dialog(bounds *b, char *desc);
	~ht_dialog();
/* overridden */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	int getstate();
	virtual	void setstate(int state);
};

/*
 *	CLASS ht_inputfield
 */

struct ht_inputfield_data {
	char *text;
};

class ht_inputfield: public ht_view {
protected:
	char *text;
	char *curchar;
	int insert;
	int maxstrlen;
	int ofs;
public:
	ht_inputfield(bounds *b, char *deftext, int maxstrlen);
	virtual ~ht_inputfield();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
/* new */
			void setchar(char a, int insert);
};

/*
 *	CLASS ht_inputfield_button
 */

struct ht_inputfield_button_data {
	char *text;
	int pressed;
};

class ht_inputfield_button: public ht_inputfield {
private:
	int value;
	int pressed;
public:
	ht_inputfield_button(bounds *b, char *deftext, int _maxstrlen, int value);
	virtual ~ht_inputfield_button();
/* overwritten */
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_button
 */

struct ht_button_data {
	int pressed;
};

class ht_button: public ht_view {
protected:
	int value;
	int pressed;
	char *text;
public:
	ht_button(bounds *b, char *text, int value);
	virtual ~ht_button();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_listbox_body
 */

struct ht_listbox_data {
	ht_string_collection *list;
	int index;
};

class ht_listbox_body: public ht_view {
protected:
	ht_string_collection *list;
	int pos, cursor;
	int vis_h_size;
	ht_scrollbar *scrollbar;
public:
	ht_listbox_body(bounds *b, ht_scrollbar *scrollbar);
	virtual ~ht_listbox_body();
/* overwritten */
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void insertstring(char *str);
			int seek(int index);
};

/*
 *	CLASS ht_listbox
 */

class ht_listbox: public ht_group {
protected:
	ht_listbox_body *body;
	ht_scrollbar *scrollbar;
public:
	ht_listbox(bounds *b);
	~ht_listbox();
/* overwritten */
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void setpalette(palette *palette);
/* new */
	void insertstring(char *str);
	int seek(int index);
};

/*
 *	CLASS ht_listbox_button
 */

class ht_listbox_button: public ht_listbox {
private:
	int value;
public:
	ht_listbox_button(bounds *b, int value);
	virtual ~ht_listbox_button();
/* overwritten */
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_listpopup_body
 */

#define ht_listpopup_data ht_listbox_data

class ht_listpopup_body: public ht_listbox_body {
protected:
	int expanded;
public:
	ht_listpopup_body(bounds *b, ht_scrollbar *scrollbar);
	virtual ~ht_listpopup_body();
/* overwritten */
			void collapse();
	virtual	void draw();
			void expand();
	virtual	void handlemsg(htmsg *msg);
	virtual	void insertstring(char *str);
	virtual	void releasefocus();
};

/*
 *	CLASS ht_listpopup
 */

class ht_listpopup: public ht_group {
protected:
	ht_listbox_body *body;
	ht_scrollbar *scrollbar;
public:
	ht_listpopup(bounds *b);
	~ht_listpopup();
/* overwritten */
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void setpalette(palette *palette);
/* new */
	void insertstring(char *str);
};

/*
 *	CLASS ht_statictext
 */

struct ht_statictext_data {
	char *text;
};

class ht_statictext: public ht_view {
private:
	char *text;
public:
	ht_statictext(bounds *b, char *text);
	virtual ~ht_statictext();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
};

/*
 *	CLASS ht_color_block
 */

struct ht_color_block_data {
	int color;
};

class ht_color_block: public ht_view {
protected:
	int color;
	int colors;
public:
	ht_color_block(bounds *b, int light_colors, int selected);
	~ht_color_block();
/* overwritten */
	virtual	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
/* new */
	virtual	void set_color(int i);
};

/*
 *	CLASS ht_listbox_ptr
 */

class ht_listbox_ptr: public ht_data {
public:
	ht_listbox *listbox;
	ht_listbox_ptr(ht_listbox *listbox);
	~ht_listbox_ptr();
};

/*
 *	CLASS ht_color_manager
 */

struct ht_color_manager_data {
	palette_group *pal_grp;
};

class ht_color_manager: public ht_dialog {
protected:
	ht_listbox *group;
	ht_xgroup *items;
	ht_collection *group_to_items;
	ht_color_block *fg_colors;
	ht_color_block *bg_colors;
	palette_group *orig_pal_grp;
	palette_group *pal_grp;
public:
	ht_color_manager(bounds *b, palette_group *pal);
	~ht_color_manager();
/* overridden */
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	misc
 */

#define msgbox_ok		1
#define msgbox_cancel	2
#define msgbox_yes		4
#define msgbox_no		8

int execute_dialog(ht_dialog *dialog);
int msgbox(int buttonmask, char *title, char *format, ...);
int warning(char *format, ...);
int error(char *format, ...);

/*
 *	INIT
 */

int init_dialog();

/*
 *	DONE
 */

int done_dialog();

#endif /* !__HTDIALOG_H__ */

