/* 
 *	The HTE Editor
 *	htfile.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTFILE_H__
#define __HTFILE_H__

#include "global.h"

#include <stdio.h>

/* some stdio.h define getc/putc macros */
#undef getc
#undef putc

/*
 *	CLASS htfile
 */

class htfile {
protected:
	int error;
	void seterror(int error);
public:
	htfile();
	virtual ~htfile();
/* new */
	ABSTRACT virtual int read(void *buf, int size);
	ABSTRACT virtual int write(void *buf, int size);
	ABSTRACT virtual int seek(int pos);
	ABSTRACT virtual int tell();
	ABSTRACT virtual int size();
	ABSTRACT virtual char *getname();
	STUB virtual void flush();
	STUB virtual void invd();
	STUB virtual int dirty();
	STUB virtual int bytedirty(int ofs);
		int geterror();
};

/*
 *	CLASS htrawfile
 */

class htrawfile: public htfile {
protected:
	char *name, *mode;
	int isdirty;
public:
	FILE *file;

	htrawfile();
	htrawfile(const char *name, const char *mode);
	virtual ~htrawfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual int size();
	virtual char *getname();
	virtual int dirty();
};

/*
 *	CLASS htrwfile
 */

#include "htdata.h"

class htrwfile: public htrawfile {
protected:
	ht_tree *vmod;
	int frag_count;
/* new */
	void putc(unsigned char c);
	unsigned char getc();
	void init_buf();
	void done_buf();
public:
	htrwfile();
	htrwfile(const char *name);
	virtual ~htrwfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual void flush();
	virtual void invd();
	virtual int bytedirty(int ofs);
};

/*
 *	CLASS htmemfile
 */

class htmemfile: public htfile {
protected:
	int ofs;
	int pos;
	int bufsize;
	char *buf;
	int isdirty;

	void extendbuf();
public:
	htmemfile();
	htmemfile(int ofs, int size);
	virtual ~htmemfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
	virtual int seek(int pos);
	virtual int tell();
	virtual int size();
	virtual char *getname();
	virtual int dirty();
};

/*
 *	string file functions
 */

char *getstrz(htfile *file);
void putstrz(htfile *file, char *str);

char *getstrl(htfile *file);
void putstrl(htfile *file, char *str);

#endif /* !__HTFILE_H__ */
