/* 
 *	The HTE Editor
 *	hthelp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htdialog.h"
#include "hthelp.h"
#include "htkeyb.h"
#include "htpal.h"
#include "httools.h"
#include "tools.h"

#include <string.h>

ht_help *help;

#define help_palette_str					"help"
#define help_palette_colors				3
#define help_coloridx_normal				0
#define help_coloridx_link_selected		1
#define help_coloridx_link_unselected		2
#define help_color_normal				pal[help_coloridx_normal].value
#define help_color_link_selected			pal[help_coloridx_link_selected].value
#define help_color_link_unselected			pal[help_coloridx_link_unselected].value

/*
 *	Info files
 */

#define INFO_MAX_LINE_LENGTH 1024

char readbuf[INFO_MAX_LINE_LENGTH];

void cleanup_str(char *s)
{
	while (*s) {
		if ((unsigned char)*s<32) *s=32;
		s++;
	}
}

void read_whitespaces(htfile *file, char *c)
{
	while (*c<=32) {
		if (!file->read(c, 1)) break;
	}
}

void read_skip(htfile *file, char *c)
{
	file->read(c, 1);
}

char *read_until(htfile *file, char *c, char *stop)
{
	char *s=readbuf;
	while (!strchr(stop, *c)) {
		*(s++)=*c;
		if (!file->read(c, 1)) break;
		if (s-readbuf>=INFO_MAX_LINE_LENGTH) break;
	}
	*s=0;
	return readbuf;
}

void read_skip_until(htfile *file, char *c, char *stop)
{
	while (!strchr(stop, *c)) {
		if (!file->read(c, 1)) break;
	}
}

int read_info_file_node(htfile *file, char *c)
{
	char *x;
	read_skip(file, c);
	read_whitespaces(file, c);
	char *filename=strdup(read_until(file, c, ","));
	read_skip(file, c);
	read_whitespaces(file, c);
	x=read_until(file, c, ":");
	if (strcmp(x, "Node")!=0) return 0;
	read_skip(file, c);
	read_whitespaces(file, c);
	char *nodename=strdup(read_until(file, c, ","));
	x=read_until(file, c, "\n");
	read_skip(file, c);
/* file already there ? */
	ht_helpfile *hf=help->findfile(filename);
	if (!hf) {
		hf=new ht_helpfile(filename);
		help->insertfile(hf);
	}

/* duplicate node ? */
	if (hf->findnode(nodename)) return 0;

/* create the node */
	ht_helpnode *hn=new ht_helpnode(nodename);
	hf->insertnode(hn);
/* read the content line by line */
	while (1) {
		if ((*c==0) || (*c==0x1f)) {
			break;
		}
		x=read_until(file, c, "\n");
		read_skip(file, c);
		cleanup_str(x);
		hn->content->add(x);
	}

	return 1;
}

int read_info_file_tags(htfile *file, char *c)
{
	read_skip_until(file, c, "\n");
	read_skip(file, c);
/* read tags */
	/* FIXME: do we need the tag table ? */
	read_skip_until(file, c, "\x1f");
	return 1;
}

int read_info_file_block(htfile *file, char *c)
{
	read_whitespaces(file, c);
	char *blocktype=read_until(file, c, ":");
//     printf("blocktype: %s\n", blocktype);
	if (strcmp(blocktype, "File")==0) {
		if (!read_info_file_node(file, c)) return 0;
	} else if (strcmp(blocktype, "Tag Table")==0) {
		if (!read_info_file_tags(file, c)) return 0;
	}
	return 1;
}

int read_info_file(htfile *file)
{
	char c=0;
	file->seek(0);
	file->read(&c, 1);
	read_skip_until(file, &c, "\x1f");
	while (1) {
		if (c==0x1f) {
			if (!read_info_file_block(file, &c)) return 0;
		} else break;
	}
	return 1;
}

/*
 *	help system
 */

void popup_help_dialog(char *file, char *node)
{
	bounds b, c;
	b.w=60;
	b.h=15;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	ht_dialog *dialog=new ht_dialog(&b, "help");

	b.x++;
	b.y++;
	b.w-=2;
	b.h-=2;
	c=b;

	b.x+=b.w;
	b.w=1;
	ht_scrollbar *scrollbar=new ht_scrollbar(&b);
	dialog->setscrollbar(scrollbar);

	dialog->setpalette(pal_grp->get(window_cyan_palette_str));

	b=c;
	ht_helpviewer *helpviewer=new ht_helpviewer(&b, scrollbar);
	dialog->insert(helpviewer);

	if (helpviewer->gotonode(file, node)) {
		execute_dialog(dialog);
	} else {
		error("help topic '(%s)%s' not found", file, node);
	}
	delete dialog;
}

/*
 *	CLASS ht_helpnode
 */

ht_helpnode::ht_helpnode(char *_name)
{
	name=strdup(_name);
	content=new ht_string_collection();
}

ht_helpnode::~ht_helpnode()
{
	delete name;
	delete content;
}

void ht_helpnode::debug_dump()
{
	printf("node: %s\n", name);
}

/*
 *	CLASS ht_helpfile
 */

ht_helpfile::ht_helpfile(char *_name)
{
	name=strdup(_name);
	nodes=new ht_collection();
}

ht_helpfile::~ht_helpfile()
{
	delete name;
	delete nodes;
}

ht_helpnode *ht_helpfile::findnode(char *name)
{
	int c=nodes->count();
	ht_helpnode *n;
	for (int i=0; i<c; i++) {
		n=(ht_helpnode*)nodes->get(i);
		if (strcmp(n->name, name)==0) return n;
	}
	return 0;
}

void ht_helpfile::insertnode(ht_helpnode *node)
{
	nodes->add(node);
}

void ht_helpfile::debug_dump()
{
	printf("file %s\n", name);
	int c=nodes->count();
	ht_helpnode *n;
	for (int i=0; i<c; i++) {
		n=(ht_helpnode*)nodes->get(i);
		n->debug_dump();
	}
}

/*
 *	CLASS ht_help
 */

ht_help::ht_help()
{
	files=new ht_collection();
}

ht_help::~ht_help()
{
	delete files;
}

ht_helpfile *ht_help::findfile(char *name)
{
	int c=files->count();
	ht_helpfile *n;
	for (int i=0; i<c; i++) {
		n=(ht_helpfile*)files->get(i);
		if (strcmp(n->name, name)==0) return n;
	}
	return 0;
}

void ht_help::insertfile(ht_helpfile *file)
{
	files->add(file);
}

void ht_help::debug_dump()
{
	printf("help dump\n");
	int c=files->count();
	ht_helpfile *n;
	for (int i=0; i<c; i++) {
		n=(ht_helpfile*)files->get(i);
		n->debug_dump();
	}
}

/*
 *	CLASS ht_helpviewer
 */

ht_helpviewer::ht_helpviewer(bounds *b, ht_scrollbar *_scrollbar)
: ht_viewer(b, "help", 0, htvc_search)
{
	scrollbar=_scrollbar;
	content=0;
	filename=0;
	cursorx=0;
	cursory=0;
	ofsx=0;
	ofsy=0;
	setpalette(pal_grp->get(help_palette_str));
}

ht_helpviewer::~ht_helpviewer()
{
}

int ht_helpviewer::cursor_up(int n)
{
	cursory-=n;
	if (cursory<0) {
		scroll_up(-cursory);
		cursory=0;
	}
	update_scrollbar();
	return n;
}

int ht_helpviewer::cursor_down(int n)
{
	int c=content->count();
	int vy=MIN(size.h-1, c);
	cursory+=n;
	if (cursory>vy) {
		scroll_down(cursory-vy);
		cursory=vy;
	}
	update_scrollbar();
	return n;
}

int ht_helpviewer::cursor_left(int n)
{
	cursorx-=n;
	if (cursorx<0) {
		scroll_left(-cursorx);
		cursorx=0;
	}
	return n;
}

int ht_helpviewer::cursor_right(int n)
{
	cursorx+=n;
	if (cursorx>size.w-1) {
		scroll_right(cursorx-size.w+1);
		cursorx=size.w-1;
	}
	return n;
}

void ht_helpviewer::cursor_home()
{
	ofsx=0;
	cursorx=0;
}

void ht_helpviewer::cursor_end()
{
	if (content) {
		cursorx=size.w-4;
		ofsx=curline_vlen-size.w+3;
		if (ofsx<0) {
			cursorx=curline_vlen-1;
			ofsx=0;
		}
	}
}

void ht_helpviewer::draw()
{
	curlink_node_dest=0;
	curlink_file_dest=0;
	int col=help_color_normal;
	buf->fill(size.x, size.y, size.w, size.h, col, ' ');
	if (content) {
		int c=content->count();
		if (c>size.h) {
			c=size.h;
		}
		for (int i=0; i<c; i++) {
			char *s=content->get(i+ofsy);
			help_print(0, i, s);
		}
	}
	setcursor(cursorx, cursory);
}

int ht_helpviewer::gotonode(char *file, char *node)
{
	ht_helpfile *hf=help->findfile(file);
	if (!hf) return 0;
	ht_helpnode *hn=hf->findnode(node);
	if (!hn) return 0;
	content=hn->content;
	if (filename) delete filename;
	filename=strdup(file);
	return 1;
}

void ht_helpviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
			case K_Space:
				if (curlink_node_dest && curlink_file_dest) gotonode(curlink_file_dest, curlink_node_dest);
				clearmsg(msg);
				dirtyview();
				break;
		}
	}
	ht_viewer::handlemsg(msg);
}

void ht_helpviewer::help_print(int x, int y, char *str)
{
	int l=strlen(str);
	int a=0;
	if (l>ofsx) {
		char *s=str, *t;
		while (1) {
			t=strchr(s, '*');
			if (!t) break;
			if (x+a<ofsx) {
				if (t-s>ofsx) {
					lprint(0, y, help_color_normal, MIN(size.w-a, t-s-ofsx), s+ofsx);
				}
			} else {
				lprint(x+a-ofsx, y, help_color_normal, MIN(size.w-a, t-s), s);
			}
			a+=t-s;
			char *v, *w;
			if (strncmp(t, "*Note ", 6)==0) {
				v=t+6;
				w=strchr(v, ':');
			} else if (strncmp(t, "* ", 2)==0) {
				v=t+2;
				w=strchr(v, ':');
			} else {
				s=t;
				continue;
			}
			if (w) {
				int cursor_here=((y==cursory) && (cursorx>=a) && (cursorx<a+w-v));
				int col=cursor_here ? help_color_link_selected : help_color_link_unselected;
				if (x+a<ofsx) {
					if (w-v>ofsx) {
						lprint(0, y, col, MIN(size.w-a, w-v-ofsx), v+ofsx);
					}
				} else {
					lprint(x+a-ofsx, y, col, MIN(size.w-a, w-v), v);
				}
				a+=w-v;
/*				lprint(x+a, y, col, MIN(size.w-a, w-v), v);
				a+=w-v;*/
				s=w+1;
				char *p=0, *d=curlink_node_dest_s;
				while (!p) {
					p=strchr(":,.", *s);
					if (!*s) break;
					s++;
					if (cursor_here) {
						*(d++)=*s;
					}
				}
				if (cursor_here) {
					if (*p!=':') {
						*(d-2)=0;
						curlink_node_dest=curlink_node_dest_s;
						strcpy(curlink_file_dest_s, filename);
						curlink_file_dest=curlink_file_dest_s;
					} else {
						strncpy(curlink_node_dest_s, v, w-v);
						curlink_node_dest=curlink_node_dest_s;
						strcpy(curlink_file_dest_s, filename);
						curlink_file_dest=curlink_file_dest_s;
					}
				}
			}
		}
		int l=strlen(s);
		if (x+a<ofsx) {
			if (l>ofsx) {
				lprint(0, y, help_color_normal, size.w-a, s+ofsx);
				a+=l-ofsx;
			}
		} else {
			lprint(x+a-ofsx, y, help_color_normal, size.w-a, s);
			a+=l;
		}
//		lprint(x+a, y, help_color_normal, size.w-a, s);
//          a+=strlen(s);
	}
	if (cursory==y) {
		curline_vlen=a+ofsx;
	}
}

void ht_helpviewer::scroll_up(int n)
{
	ofsy-=n;
	if (ofsy<0) ofsy=0;
}

void ht_helpviewer::scroll_down(int n)
{
	int c=content->count();
	int vy=MIN(size.h-1, c);
	ofsy+=n;
	if (ofsy+vy>c) {
		ofsy=c-vy;
	}
}

void ht_helpviewer::scroll_left(int n)
{
	ofsx-=n;
	if (ofsx<0) {
		ofsx=0;
	}
}

void ht_helpviewer::scroll_right(int n)
{
	ofsx+=n;
}

void ht_helpviewer::update_scrollbar()
{
	if (content) {
		int c=content->count();
		if (c>1) scrollbar->setpos((ofsy+cursory)*100/(c-1));
	}
}

/*
 *	INIT
 */

void init_help()
{
	palette *help_palette=pal_grp->create(help_palette_str, help_palette_colors);
	pal_grp->set_entry(help_palette, help_coloridx_normal, "normal", COLOR(C_BLACK, C_CYAN));
	pal_grp->set_entry(help_palette, help_coloridx_link_selected, "selected link", COLOR(C_LIGHT(C_WHITE), C_BLUE));
	pal_grp->set_entry(help_palette, help_coloridx_link_unselected, "unselected link", COLOR(C_LIGHT(C_WHITE), C_CYAN));

	help=new ht_help();
}

/*
 *	DONE
 */

void done_help()
{
	delete help;
}

