/* 
 *	The HTE Editor
 *	hthelp.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTHELP_H__
#define __HTHELP_H__

#include "common.h"
#include "htdata.h"
#include "htfile.h"
#include "htobj.h"

int read_info_file(htfile *file);
void popup_help_dialog(char *file, char *node);

/*
 *	CLASS ht_helpnode
 */

class ht_helpnode: public ht_data {
public:
	char *name;
	ht_string_collection *content;

	ht_helpnode(char *name);
	virtual ~ht_helpnode();
/* debug */
	void debug_dump();
};

/*
 *	CLASS ht_helpfile
 */

class ht_helpfile: public ht_data {
protected:
	ht_collection *nodes;
public:
	char *name;

	ht_helpfile(char *name);
	virtual ~ht_helpfile();
/* new */
	ht_helpnode *findnode(char *name);
	void insertnode(ht_helpnode *node);
/* debug */
	void debug_dump();
};

/*
 *	CLASS ht_help
 */

class ht_help: public object {
public:
	ht_collection *files;

	ht_help();
	virtual ~ht_help();
/* new */
	ht_helpfile *findfile(char *name);
	void insertfile(ht_helpfile *file);
/* debug */
	void debug_dump();
};

/*
 *	CLASS ht_helpviewer
 */

class ht_helpviewer: public ht_viewer {
protected:
	ht_string_collection *content;
	int cursorx, cursory;
	int ofsx, ofsy;
	ht_scrollbar *scrollbar;
	char curlink_node_dest_s[128], curlink_file_dest_s[128];
	char *curlink_node_dest, *curlink_file_dest;
	char *filename;
	int curline_vlen;
public:
	ht_helpviewer(bounds *b, ht_scrollbar *scrollbar);
	virtual ~ht_helpviewer();
/* overwritten */
	virtual int cursor_up(int n);
	virtual int cursor_down(int n);
	virtual int cursor_left(int n);
	virtual int cursor_right(int n);
		   void cursor_home();
		   void cursor_end();
	virtual void draw();
	virtual void handlemsg(htmsg *msg);
	virtual void scroll_up(int n);
	virtual void scroll_down(int n);
	virtual void scroll_left(int n);
	virtual void scroll_right(int n);
/* new */
		   int gotonode(char *file, char *node);
		   void help_print(int x, int y, char *str);
		   void update_scrollbar();
};

/*
 *	INIT
 */

void init_help();

/*
 *	DONE
 */

void done_help();

#endif  /* !__HTHELP_H__ */
