/* 
 *	The HTE Editor
 *	htobj.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __HTOBJ_H__
#define __HTOBJ_H__

class ht_view;
class ht_group;

#include "common.h"
#include "htcurses.h"
#include "htdata.h"
#include "htdebug.h"
#include "htexpl.h"
#include "htfile.h"
#include "htpal.h"
#include "x86dis.h"
#include "viewers.h"

#define htm_nothing 		0
#define htm_draw 			1
#define htm_resize 			2
#define htm_keypressed 		3
#define htm_complete		4
#define htm_funcexec		5
#define htm_funcname		6
#define htm_funcname_return	7
#define htm_button_pressed	8
#define htm_dirtyview		9

#define htmt_nothing		0
#define htmt_broadcast		1

struct htmsg {
	int msg;
	int type;
	union {
		int data_int;
		void *data_ptr;
		char *data_str;
	};
};

/*
 *	Palette entries
 */

#define window_cyan_palette_str		"window (cyan)"
#define window_black_palette_str		"window (black)"

#define window_palette_str		window_black_palette_str
#define window_palette_colors		4
#define window_coloridx_normal	0
#define window_coloridx_frame		1
#define window_coloridx_scrollbar	2
#define window_coloridx_killer	3
#define window_color_normal		pal[window_coloridx_normal].value
#define window_color_frame		pal[window_coloridx_frame].value
#define window_color_scrollbar	pal[window_coloridx_scrollbar].value
#define window_color_killer   	pal[window_coloridx_killer].value

#define tag_palette_str			"tags"
#define tag_palette_colors		7
#define tag_color_normal			pal[tag_coloridx_normal].value
#define tag_color_edit_selected	pal[tag_coloridx_edit_selected].value
#define tag_color_edit_unselected	pal[tag_coloridx_edit_unselected].value
#define tag_color_edit_edited		pal[tag_coloridx_edit_edited].value
#define tag_color_edit_modified	pal[tag_coloridx_edit_modified].value
#define tag_color_sel_selected	pal[tag_coloridx_sel_selected].value
#define tag_color_sel_unselected	pal[tag_coloridx_sel_unselected].value

/*
 *	CLASS ht_view
 */

#define hto_ownbuffer		1
#define hto_browsable		2
#define hto_selectable		4

class ht_group;

class ht_view: public object {
protected:
	int focused;
	int view_is_dirty;
	bounds size;
	palette *pal;

	void cleanview();
	void dirtyview();
public:
	int enabled;
	genericdrawbuf *buf;
	int options;
	char *desc;
	ht_group	*group;
	int browse_idx;
	ht_view 	*prev, *next;

			ht_view(bounds *b, int options, char *desc);
	virtual 	~ht_view();
/* new */
	virtual	int childcount();
			void clear();
			void clearmsg(htmsg *msg);
	virtual	void disable();
			void disable_buffering();
STUB	virtual 	void draw();
	virtual	void enable();
			void enable_buffering();
	virtual	int focus(ht_view *view);
			void getbounds(bounds *b);
	virtual	ht_view *getcurrentchild();
	virtual 	ht_view *getfirstchild();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	int getnumber();
	virtual 	void handlemsg(htmsg *msg);
			void hidecursor();
			int isviewdirty();
			int lprint(int x, int y, int c, int l, char *text);
			void move(int x, int y);
			int printf(int x, int y, int c, char *format, ...);
			int print(int x, int y, int c, char *text);
			void printchar(int x, int y, int c, int ch);
	virtual	void receivefocus();
			void redraw();
	virtual 	void releasefocus();
	virtual	int select(ht_view *view);
			void sendmsg(int msg, int data);
			void sendmsg(htmsg *msg);
			void setbounds(bounds *b);
			void setcursor(int x, int y);
	virtual	void setpalette(palette *palette);
};

/*
 *	CLASS ht_group
 */

class ht_group: public ht_view {
protected:
	ht_view 	*first, *current, *last;
	int 		view_count;
public:
	void		*shared_data;


			ht_group(bounds *b, int options, char *desc);
	virtual	~ht_group();
/* overwritten */
	virtual	int childcount();
	virtual	int focus(ht_view *view);
	virtual	ht_view *getcurrentchild();
	virtual 	ht_view *getfirstchild();
STUB	virtual	void handlemsg(htmsg *msg);
			int isviewdirty();
			void putontop(ht_view *view);
	virtual 	void receivefocus();
			void redraw();
	virtual 	void releasefocus();
	virtual	int select(ht_view *view);
/* new */
			void insert(ht_view *view);
			int focusnext();
			int focusprev();
			ht_view *get_by_browse_idx(int i);
};

/*
 *	CLASS ht_xgroup
 */

class ht_xgroup: public ht_group {
public:
			ht_xgroup(bounds *b, int options, char *desc);
	virtual	~ht_xgroup();
/* overwritten */
STUB	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_scrollbar
 */

class ht_scrollbar: public ht_view {
protected:
	int ppos;
public:
	ht_scrollbar(bounds *b);
	~ht_scrollbar();
/* overridden */
	virtual	void enable();
	virtual	void disable();
	virtual 	void draw();
/* new */
	virtual	void setpos(int percent);
};

/*
 *	CLASS ht_winframe
 */

class ht_winframe: public ht_view {
protected:
	int number;
public:
			ht_winframe(bounds *b, char *desc, int number);
	virtual 	~ht_winframe();
/* overwritten */
STUB	virtual	void draw();
STUB	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_window
 */

class ht_window: public ht_group {
protected:
	ht_winframe *frame;
	ht_scrollbar *scrollbar;
	int number;
public:
			ht_window(bounds *b, char *desc, int number);
	virtual 	~ht_window();
/* overwritten */
	virtual	void draw();
	virtual	int getnumber();
	virtual	void setpalette(palette *palette);
/* new */
			void setscrollbar(ht_scrollbar *scrollbar);
};

/*
 *	CLASS ht_viewer
 */

#define htvc_edit			1
#define htvc_goto			2
#define htvc_search			4

class ht_viewer: public ht_view {
protected:
	htfile *file;
	int caps;
public:
	void *shared_data;					/* data being shared with childs */

	ht_viewer(bounds *b, char *desc, htfile *file, int caps);
	virtual ~ht_viewer();
/* overwritten */
	STUB virtual void handlemsg(htmsg *msg);
/* new */
	STUB virtual void complete();
	STUB virtual int cursor_up(int n);
	STUB virtual int cursor_down(int n);
	STUB virtual int cursor_left(int n);
	STUB virtual int cursor_right(int n);
	STUB virtual void cursor_home();
	STUB virtual void cursor_end();
	STUB virtual char *func(int i, int getname);
	STUB virtual int jump(ADDR ofs);
	STUB virtual int jumpstr(char *str);
	STUB virtual void scroll_up(int n);
	STUB virtual void scroll_down(int n);
	STUB virtual void scroll_left(int n);
	STUB virtual void scroll_right(int n);
};

/*
 *	CLASS ht_uviewer
 */

class hts_sub; /* to resolve forward ref */

class ht_uviewer: public ht_viewer {
protected:
	hts_sub *firstsub, *currentsub, *lastsub, *cursor_sub;
	ADDR ofs;
	int yofs, xofs;
	ADDR cursor_eofs;
	ADDR cursor_ofs;
	int cursor_yofs;
	int cursor_e, cursor_x, cursor_y, cursor_group;
	int is_sel_tag;
	char curline[1024];
	int edit;
/* new */
	void update_cursor_data();
	int seek_prev_line(hts_sub **sub, ADDR *ofs, int *yofs, int n);
	int seek_next_line(hts_sub **sub, ADDR *ofs, int *yofs, int n);
	int seek_prev_editable_line(hts_sub **sub, ADDR *ofs, int *yofs);
	int seek_next_editable_line(hts_sub **sub, ADDR *ofs, int *yofs);
	int count_lines_to_top(hts_sub *sub, ADDR ofs, int yofs);
	int count_lines_to_bottom(hts_sub *sub, ADDR ofs, int yofs);
	void apply_modifier(char *htstring);
	void cursor_next_group();
	int start_edit();
	int update_edit();
	int stop_edit();
	void qdraw();
	int editinput(int key);
	void popup_jump_dialog();
	void popup_search_dialog();
public:
		ht_uviewer(bounds *b, char *desc, htfile *file, int caps);
		virtual ~ht_uviewer();
/* overwritten */
		virtual void draw();
		virtual void handlemsg(htmsg *msg);
		virtual void complete();
		virtual int cursor_up(int n);
		virtual int cursor_down(int n);
		virtual int cursor_left(int n);
		virtual int cursor_right(int n);
		virtual void cursor_home();
		virtual void cursor_end();
		virtual char *func(int i, int getname);
		virtual int jump(ADDR ofs);
		virtual void ref(dword id);
		virtual void scroll_up(int n);
		virtual void scroll_down(int n);
		virtual void scroll_left(int n);
		virtual void scroll_right(int n);
/* new */
		virtual void insertsub(hts_sub *sub);
			   int htprint(int x, int y, int maxlen, int ofs, char *text);
};

/*
 *	CLASS ht_logviewer
 */

class ht_logviewer: public ht_viewer {
private:
	ht_string_collection *lines;
	int maxlinecount;
	int ofs, xofs;

	void deletefirstline();
public:
	ht_logviewer(bounds *b, int _maxlinecount);
	~ht_logviewer();
/* overwritten */
	virtual void draw();
	virtual void handlemsg(htmsg *msg);
/* new */
	void insertline(char *_line);
};

/*
 *	CLASS hts_sub
 */

class hts_sub: public object {
protected:
	htfile *file;
public:
	hts_sub *prev, *next;

	hts_sub(htfile *file);
	virtual ~hts_sub();
/* new */
	ABSTRACT virtual int getline(char *line, ADDR ofs, int yofs);
	ABSTRACT virtual void start_line_address(ADDR *ofs, int *yofs);
	ABSTRACT virtual void end_line_address(ADDR *ofs, int *yofs);
	ABSTRACT virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
	ABSTRACT virtual int next_line_address(ADDR *ofs, int *yofs, int n);
	STUB virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_hex
 */

class hts_hex: public hts_sub {
private:
	ADDR fofs;
	dword fsize;
public:
	hts_hex(htfile *file, int ofs, int size);
	virtual ~hts_hex();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
	virtual void start_line_address(ADDR *ofs, int *yofs);
	virtual void end_line_address(ADDR *ofs, int *yofs);
	virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
	virtual int next_line_address(ADDR *ofs, int *yofs, int n);
	virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_unasm
 */

class hts_unasm: public hts_sub {
private:
	ADDR fofs;
	dword fsize;
	x86dis *unasm;
public:
	hts_unasm(htfile *file, ADDR ofs, int size);
	virtual ~hts_unasm();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
	virtual void start_line_address(ADDR *ofs, int *yofs);
	virtual void end_line_address(ADDR *ofs, int *yofs);
	virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
	virtual int next_line_address(ADDR *ofs, int *yofs, int n);
	virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_text
 */

class hts_text: public hts_sub {
private:
	char *text;
public:
	hts_text(htfile *file, const char *text);
	virtual ~hts_text();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
	virtual void start_line_address(ADDR *ofs, int *yofs);
	virtual void end_line_address(ADDR *ofs, int *yofs);
	virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
	virtual int next_line_address(ADDR *ofs, int *yofs, int n);
};

/*
 *	CLASS hts_mask
 */

class hts_mask: public hts_sub {
private:
	ht_string_collection masks;
	char temp[512];
	int start;
public:
	hts_mask(htfile *file, int start);
	virtual ~hts_mask();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
	virtual void start_line_address(ADDR *ofs, int *yofs);
	virtual void end_line_address(ADDR *ofs, int *yofs);
	virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
	virtual int next_line_address(ADDR *ofs, int *yofs, int n);
/* new */
	virtual void add_mask(char *htstring);
	virtual void add_mask_table(char **htstring);
};

/*
 *	INIT
 */

int init_obj();

/*
 *	DONE
 */

int done_obj();

#endif /* !__HTOBJ_H__ */
