/* 
 *	The HTE Editor
 *	htpal.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "httools.h"
#include "htpal.h"

#include <stdlib.h>
#include <string.h>

palette_group *pal_grp;

/*
 *	CLASS palette_group
 */

palette_group::palette_group()
{
	pal=0;
	enum_pal=0;
}

palette_group::~palette_group()
{
	freeall();
}

palette_group *palette_group::copy()
{
	palette_group *g=new palette_group();
	ipalette *p=pal;
	while (p) {
		palette *h=g->create(p->name, p->entries);
		for (int i=0; i<p->entries; i++) {
			g->set_entry(h, i, (p->pal+i)->name, (p->pal+i)->value);
		}
		p=p->next;
	};
	return g;
}

palette *palette_group::create(char *name, int entries)
{
	ipalette *p=icreate(name, entries);
	if (p) return p->pal; else return 0;
}

palette *palette_group::create_copy(char *srcname, char *destname)
{
	ipalette *p=icreate_copy(srcname, destname);
	if (p) return p->pal; else return 0;
}

palette *palette_group::create_xcopy(char *srcname, char *destname, int entries)
{
	ipalette *p=icreate_xcopy(srcname, destname, entries);
	if (p) return p->pal; else return 0;
}

ipalette *palette_group::enum_first()
{
	enum_pal=pal;
	return pal;
}

ipalette *palette_group::enum_next()
{
	if (enum_pal) enum_pal=enum_pal->next;
	return enum_pal;
}

palette *palette_group::get(char *name)
{
	ipalette *p=iget(name);
	if (p) return p->pal; else return 0;
}

int palette_group::get_color(int fg, int bg)
{
	return ((bg&7)<<4)|(fg&0xf);
}

void palette_group::freeall()
{
	ipalette *p=pal;
	while (p) {
		ipalette *q=p->next;
		delete p;
		p=q;
	}
	pal=0;
	enum_pal=0;
}

ipalette *palette_group::icreate(char *name, int entries)
{
	if (get(name)) return 0;
	ipalette *p=new ipalette;
	p->next=pal;
	p->name=name;
	p->entries=entries;
	p->pal=(palette*)malloc(entries * sizeof *(p->pal));
	pal=p;
	return p;
}

ipalette *palette_group::icreate_copy(char *srcname, char *destname)
{
	ipalette *src=iget(srcname);
	if (!src) return 0;
	ipalette *dest=icreate(destname, src->entries);
	if (!dest) return 0;
	memmove(dest->pal, src->pal, src->entries * sizeof *(dest->pal));
	return dest;
}

ipalette *palette_group::icreate_xcopy(char *srcname, char *destname, int entries)
{
	ipalette *src=iget(srcname);
	if (!src) return 0;
	ipalette *dest=icreate(destname, entries);
	if (!dest) return 0;
	for (int i=0; i<entries; i++) {
		dest->pal[i].name="";
		dest->pal[i].value=0;
	}
	memmove(dest->pal, src->pal, MIN(src->entries, entries) * sizeof *(dest->pal));
	return dest;
}

ipalette *palette_group::iget(char *name)
{
	ipalette *p=pal;
	while (p) {
		if (strcmp(name, p->name)==0) return p;
		p=p->next;
	}
	return 0;
}

int palette_group::get_entry(palette *palette, int entry, char **name, char *value)
{
	int max_es=-1;
	ipalette *p=pal;
	while (p) {
		if (p->pal==palette) {
			max_es=p->entries;
			break;
		}
		p=p->next;
	}
	if (entry<max_es) {
		*name=palette[entry].name;
		*value=palette[entry].value;
		return 1;
	}
	return 0;
}

void palette_group::paste(palette_group *pg)
{
	ipalette *p=pal;
	while (p) {
		ipalette *q=pg->iget(p->name);
		int c=MIN(q->entries, p->entries);
		memmove(p->pal, q->pal, c * sizeof *p->pal);
		p=p->next;
	}
}

int palette_group::set_entry(palette *palette, int entry, char *name, char value)
{
	int max_es=-1;
	ipalette *p=pal;
	while (p) {
		if (p->pal==palette) {
			max_es=p->entries;
			break;
		}
		p=p->next;
	}
	if (entry<max_es) {
		palette[entry].name=name;
		palette[entry].value=value;
		return 1;
	}
	return 0;
}

