/* 
 *	The HTE Editor
 *	htpeimp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htfile.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpeimp.h"
#include "htstring.h"
#include "viewers.h"

#include <stdlib.h>

int htpeimports_detect(htfile *file)
{
	int h=get_newexe_header_ofs(file);
	dword pe=0;
	file->seek(h);
	file->read(&pe, 2);
	if (pe==PE_MAGIC) {
		int e, f;
		file->seek(h+0x78+1*8);
		file->read(&e, 4);
		file->read(&f, 4);
		return e|f;
	}
	return 0;
}

ht_view *htpeimports_init(bounds *b, htfile *file, ht_group *group)
{
	int h0=new_timer();
	start_timer(h0);

	ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)group)->shared_data;
	int h=get_newexe_header_ofs(file);
	char iline[1024]; /* FIXME: possible buffer overflow */
	ht_group *g;
	bounds c;

	g=new ht_group(&c, 0, "pe/imports-g");
	ht_statictext *head;

	int dll_count=0;
	int function_count=0;

	c=*b;

	c.y++;
	c.h--;
	ht_uviewer *v=new ht_uviewer(&c, "pe/imports", file, 0);
	hts_mask *m=new hts_mask(file, h);

	c.y--;
	c.h=1;

/* get import directory offset */
	/* 1. get import directory rva */
	dword iofs, irva=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_IMPORT].address, isize=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_IMPORT].size;
	/* 2. transform it into an offset */
	if (!pe_rva_to_ofs(&pe_shared->sections, irva, &iofs)) goto pe_read_error;
	LOG("%s: PE: reading import directory at offset %08x, rva %08x, size %08x...", file->getname(), iofs, irva, isize);
/* make a memfile out of the whole file */

/* doesn't work because import information is NOT contained into
   the import directory !!! only the import dir header. (I !love M$) */

/*     htfile *ifile;
	char *isectionbuf;
	ifile=new htmemfile(iofs, isize);
	isectionbuf=(char*)malloc(isize);
	file->seek(iofs);
	file->read(isectionbuf, isize);
	ifile->write(isectionbuf, isize);
	file=ifile;
	delete isectionbuf;*/

/*** read import directory ***/
	PE_IMPORT_DESCRIPTOR import;

	int dofs;
	dofs=iofs;
	int f;
	f=0;
	int dll_index;
	dll_index=0;

	while (1) {
		file->seek(dofs);
		file->read(&import, sizeof import);
		if ((!import.characteristics) && (!import.name)) break;
		dofs=file->tell();
/* get name of dll */
		dword iname_ofs;
		if (!pe_rva_to_ofs(&pe_shared->sections, import.name, &iname_ofs)) goto pe_read_error;
		file->seek(iname_ofs);
		char *dllname=getstrz(file);	/* dont forget to free it at the end of the scope !!! */
		dll_count++;

/*** imported functions by name or by ordinal ***/

/*
 *	First thunk (FT)
 *   The first thunk table is what will get overwritten in run-time with
 *	the function entry-points. So the program will refer to this table
 *	as if it contained them.
 */
		dword fthunk_rva=import.first_thunk, fthunk_ofs;
		if (!pe_rva_to_ofs(&pe_shared->sections, fthunk_rva, &fthunk_ofs)) goto pe_read_error;
/*
 *   FT vs original FT
 * 	I saw executables that have the original FT ptr set to 0 and seem to
 *	use the FT ptr instead, but also some that have both !=0 and use
 *	the original one.
 * 	I don't know if this is right... please tell me if you find out ;)
 */
		dword thunk_ofs, thunk_rva;
		if (import.original_first_thunk) {
			thunk_rva=import.original_first_thunk;
			if (!pe_rva_to_ofs(&pe_shared->sections, thunk_rva, &thunk_ofs)) goto pe_read_error;
		} else {
			thunk_rva=fthunk_rva;
			thunk_ofs=fthunk_ofs;
		}
		ht_pe_import_library *lib=new ht_pe_import_library(dllname);
		pe_shared->imports.libs.add(lib);

		PE_THUNK_DATA thunk;
		dword thunk_count=0;
		file->seek(thunk_ofs);
		while (1) {
			file->read(&thunk, sizeof thunk);
			if (!thunk.ordinal)	break;
			thunk_count++;
		}

		PE_THUNK_DATA *thunk_table;
		thunk_table=(PE_THUNK_DATA*)malloc(sizeof *thunk_table * thunk_count);
		file->seek(thunk_ofs);
		file->read(thunk_table, sizeof *thunk_table * thunk_count);

		for (dword i=0; i<thunk_count; i++) {
			function_count++;
			ht_pe_import_function *func;
/* follow (original) first thunk */
			thunk=*(thunk_table+i);

			if (thunk.ordinal & 0x80000000) {
/* ordinal */
				sprintf(iline, "%-14s %08lx "_SEL("%08x", "04")"%04lx", dllname, fthunk_rva, f, thunk.ordinal&0xffff);
				func=new ht_pe_import_function(dll_index, fthunk_rva+pe_shared->pe32header.image_base, thunk.ordinal&0xffff);
			} else {
/* name and hint */
				dword function_desc_ofs;
				int hint=0;
				if (!pe_rva_to_ofs(&pe_shared->sections, thunk.function_desc_address, &function_desc_ofs)) goto pe_read_error;
				file->seek(function_desc_ofs);
				file->read(&hint, 2);
				char *name=getstrz(file);
/* FIXME: what about the hint ? */
				sprintf(iline, "%-14s %08lx "_SEL("%08x", "%02lx")"%s", dllname, fthunk_rva, f, strlen(name), name);
				func=new ht_pe_import_function(dll_index, fthunk_rva+pe_shared->pe32header.image_base, name, hint);
				delete name;
			}
			pe_shared->imports.funcs.add(func);
			m->add_mask(iline);

			thunk_ofs+=4;
			thunk_rva+=4;
			fthunk_ofs+=4;
			fthunk_rva+=4;
			f++;
		}
		
		dll_index++;

		delete thunk_table;

		delete dllname;
	}

	sprintf(iline, "Library        FT_RVA   Name/Ordinal    %d imports from %d libraries", function_count, dll_count);
	head=new ht_statictext(&c, iline);

	if (!function_count) m->add_mask(_SEL("ffffffff", "17")"<no imported functions>");

	stop_timer(h0);
	LOG("%s: PE: it took %d ticks (%d msec) to read the imports", file->getname(), get_timer_tick(h0), get_timer_msec(h0));
	delete_timer(h0);

	v->insertsub(m);

	g->insert(head);
	g->insert(v);

	return g;
pe_read_error:
	delete m;
	delete v;
/* FIXME: return of zero not yet supported...(inits must be successful ;-) */
	assert(0);
	return 0;
}

viewer_int htpeimports_int = {
	htpeimports_detect,
	htpeimports_init
};

/*
 *	class ht_pe_import_library
 */

ht_pe_import_library::ht_pe_import_library()
{
}

ht_pe_import_library::ht_pe_import_library(char *name)
{
	this->name=htstrdup(name);
}

ht_pe_import_library::~ht_pe_import_library()
{
}

/*
 *	class ht_pe_import_function
 */

ht_pe_import_function::ht_pe_import_function()
{
	dll=0;
	name.name=0;
	name.hint=0;
	ordinal=0;
	function_address=0;
}

ht_pe_import_function::ht_pe_import_function(int dll, int function_address, int ordinal)
{
	this->dll=dll;
	this->ordinal=ordinal;
	this->function_address=function_address;
	byname=0;
}

ht_pe_import_function::ht_pe_import_function(int dll, int function_address, char *name, int hint)
{
	this->dll=dll;
	this->name.name=htstrdup(name);
	this->name.hint=hint;
	this->function_address=function_address;
	byname=1;
}

ht_pe_import_function::~ht_pe_import_function()
{
	if ((byname) && (name.name)) delete name.name;
}
