/*
 *   htcurses.h - LINUX implementation
 */

#ifndef __HTCURSES_H__
#define __HTCURSES_H__

#include "config.h"

#include CURSES_HDR
#undef clear
#undef move

#include "common.h"
#include <stdarg.h>

#define C_BLACK 0
#define C_BLUE 1
#define C_GREEN 2
#define C_CYAN 3
#define C_RED 4
#define C_MAGENTA 5
#define C_YELLOW 6
#define C_WHITE 7

#define C_LIGHT(c) (c+8)

#define COLOR(fg, bg) ((bg<<4)|fg)

#define CHAR_LINEV ACS_VLINE
#define CHAR_LINEH ACS_HLINE
#define CHAR_BORDERTL ACS_LTEE
#define CHAR_BORDERTR ACS_RTEE
#define CHAR_BORDERTU ACS_TTEE
#define CHAR_BORDERTD ACS_BTEE
#define CHAR_CORNERUL ACS_ULCORNER
#define CHAR_CORNERLL ACS_LLCORNER
#define CHAR_CORNERUR ACS_URCORNER
#define CHAR_CORNERLR ACS_LRCORNER
#define CHAR_FILLED_S 'x'
#define CHAR_FILLED_M 'x'
#define CHAR_FILLED_H 'x'
#define CHAR_QUAD_SMALL 'x'
#define CHAR_ARROW_UP 'x'
#define CHAR_ARROW_DOWN 'x'
#define CHAR_ARROWBIG_UP 'x'
#define CHAR_ARROWBIG_DOWN 'x'

struct drawbufch {
	int ch;
	int c;
};

/*
 *	CLASS genericdrawbuf
 */

class genericdrawbuf {
public:
	bounds size;

	genericdrawbuf();
	virtual ~genericdrawbuf();
	virtual void fill(int x, int y, int w, int h, int c, int ch);
		   int printf(int x, int y, int c, char *format, ...);
	virtual int print(int x, int y, int c, char *text);
	virtual void printchar(int x, int y, int c, int ch);
	virtual int lprint(int x, int y, int c, int l, char *text);
	virtual void setbounds(bounds *b);
};

/*
 *	CLASS drawbuf
 */

class drawbuf: public genericdrawbuf {
public:
	drawbufch *buf;

	drawbuf(bounds *b);
	~drawbuf();
	void fill(int x, int y, int w, int h, int c, int ch);
	int print(int x, int y, int c, char *text);
	void printchar(int x, int y, int c, int ch);
	int lprint(int x, int y, int c, int l, char *text);
	void setbounds(bounds *b);
};

class drawbuf;

class screendrawbuf: public genericdrawbuf {
public:
	screendrawbuf();
	~screendrawbuf();
	void fill(int x, int y, int w, int h, int c, int ch);
	int print(int x, int y, int c, char *text);
	void printchar(int x, int y, int c, int ch);
	int lprint(int x, int y, int c, int l, char *text);
	void setbounds(bounds *b);
/* new */
	void drawbuffer(drawbuf *buf, int x, int y);
	void show();
	void getcursor(int *x, int *y);
	void hidecursor();
	void showcursor();
	void setcursor(int x, int y);
};

extern int xres, yres;

#endif /* !__HTCURSES_H__ */
