/* 
 *	The HTE Editor
 *	pestruct.h
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __PESTRUCT_H_
#define __PESTRUCT_H_

#include "global.h"
#include "httools.h"

#include <time.h>

#define PE_MAGIC MAGICD("PE\0\0")

typedef struct	PE_HEADER {
	dword magic __attribute__ ((packed));
	word machine __attribute__ ((packed));
	word section_count __attribute__ ((packed));
	time_t timestamp __attribute__ ((packed));
	dword symbol_table_offset __attribute__ ((packed));
	dword symbol_count __attribute__ ((packed));
	word optional_header_size __attribute__ ((packed));
	word characteristics __attribute__ ((packed));
};

#define PE_SIZEOF_FILE_HEADER    		20

#define PE_RELOCS_STRIPPED		   	0x0001  // Relocation info stripped from file.
#define PE_EXECUTABLE_IMAGE		   	0x0002  // File is executable  (i.e. no unresolved externel references).
#define PE_LINE_NUMS_STRIPPED	   		0x0004  // Line nunbers stripped from file.
#define PE_LOCAL_SYMS_STRIPPED	   	0x0008  // Local symbols stripped from file.
#define PE_AGGRESIVE_WS_TRIM		   	0x0010  // Agressively trim working set
#define PE_byteS_REVERSED_LO		   	0x0080  // bytes of machine word are reversed.
#define PE_32BIT_MACHINE		   		0x0100  // 32 bit word machine.
#define PE_DEBUG_STRIPPED		   	0x0200  // Debugging info stripped from file in .DBG file
#define PE_REMOVABLE_RUN_FROM_SWAP   	0x0400  // If Image is on removable media, copy and run from the swap file.
#define PE_NET_RUN_FROM_SWAP		   	0x0800  // If Image is on Net, copy and run from the swap file.
#define PE_SYSTEM				   	0x1000  // System File.
#define PE_DLL				   		0x2000  // File is a DLL.
#define PE_UP_SYSTEM_ONLY		   	0x4000  // File should only be run on a UP machine
#define PE_BYTES_REVERSED_HI		   	0x8000  // bytes of machine word are reversed.

#define PE_MACHINE_UNKNOWN		   	0
#define PE_MACHINE_I386			   	0x14c   // Intel 386.
#define PE_MACHINE_R3000		   		0x162   // MIPS little-endian, 0x160 big-endian
#define PE_MACHINE_R4000		   		0x166   // MIPS little-endian
#define PE_MACHINE_R10000		   	0x168   // MIPS little-endian
#define PE_MACHINE_ALPHA		   		0x184   // Alpha_AXP
#define PE_MACHINE_POWERPC		   	0x1F0   // IBM PowerPC Little-Endian

//
// Directory format.
//

typedef struct	PE_DATA_DIRECTORY {
    dword	  address __attribute__ ((packed));
    dword	  size __attribute__ ((packed));
};

#define PE_NUMBEROF_DIRECTORY_ENTRIES	  	16

//
// Optional header format.
//

typedef struct	PE_OPTIONAL_HEADER	{
	word magic __attribute__ ((packed));
	byte major_linker_version __attribute__	((packed));
	byte minor_linker_version __attribute__	((packed));
	dword code_size __attribute__ ((packed));
	dword data_size __attribute__ ((packed));
	dword bss_size __attribute__ ((packed));
	dword entrypoint_address __attribute__ ((packed));
	dword code_base __attribute__ ((packed));
	dword data_base __attribute__ ((packed));

// NT additional fields.
	dword image_base __attribute__ ((packed));
	dword section_alignment __attribute__ ((packed));
	dword file_alignment __attribute__	((packed));
	word major_os_version	__attribute__ ((packed));
	word minor_os_version	__attribute__ ((packed));
	word major_image_version	__attribute__ ((packed));
	word minor_image_version	__attribute__ ((packed));
	word major_subsystem_version __attribute__ ((packed));
	word minor_subsystem_version __attribute__ ((packed));
	dword win32_version __attribute__ ((packed));
	dword image_size __attribute__ ((packed));
	dword header_size __attribute__ ((packed));
	dword checksum __attribute__	((packed));
	word	subsystem __attribute__ ((packed));
	word dll_characteristics __attribute__ ((packed));
	dword stack_reserve_size __attribute__ ((packed));
	dword stack_commit_size __attribute__ ((packed));
	dword heap_reserve_size __attribute__ ((packed));
	dword heap_commit_size __attribute__ ((packed));
	dword loader_flags __attribute__ ((packed));
	dword directory_count __attribute__ ((packed));
	PE_DATA_DIRECTORY directory[PE_NUMBEROF_DIRECTORY_ENTRIES] __attribute__ ((packed));
};

// Subsystem Values

#define PE_SUBSYSTEM_NATIVE			   1	  // Image doesn't require a subsystem.
#define PE_SUBSYSTEM_WINDOWS_GUI		   2	  // Image runs in the Windows GUI subsystem.
#define PE_SUBSYSTEM_WINDOWS_CUI		   3	  // Image runs in the Windows character subsystem.
#define PE_SUBSYSTEM_OS2_CUI			   5	  // image runs in the OS/2 character subsystem.
#define PE_SUBSYSTEM_POSIX_CUI		   7	  // image run  in the Posix character subsystem.
#define PE_SUBSYSTEM_RESERVED8		   8	  // image run  in the 8 subsystem.


// Directory Entries

#define PE_DIRECTORY_ENTRY_EXPORT			 0	  // Export Directory
#define PE_DIRECTORY_ENTRY_IMPORT			 1	  // Import Directory
#define PE_DIRECTORY_ENTRY_RESOURCE		 2	  // Resource Directory
#define PE_DIRECTORY_ENTRY_EXCEPTION		 3	  // Exception Directory
#define PE_DIRECTORY_ENTRY_SECURITY		 4	  // Security Directory
#define PE_DIRECTORY_ENTRY_BASERELOC		 5	  // Base Relocation Table
#define PE_DIRECTORY_ENTRY_DEBUG			 6	  // Debug Directory
#define PE_DIRECTORY_ENTRY_COPYRIGHT		 7	  // Description String
#define PE_DIRECTORY_ENTRY_GLOBALPTR		 8	  // Machine Value (MIPS GP)
#define PE_DIRECTORY_ENTRY_TLS			 9	  // TLS Directory
#define PE_DIRECTORY_ENTRY_LOAD_CONFIG		10	  // Load Configuration Directory
#define PE_DIRECTORY_ENTRY_BOUND_IMPORT		11	  // Bound Import Directory in headers
#define PE_DIRECTORY_ENTRY_IAT			12	  // Import Address Table
#define PE_DIRECTORY_ENTRY_DELAY_IMPORT		13	  // Delay Import Directory
#define PE_DIRECTORY_ENTRY_COM			14	  // COM+

/*
 *	Section header
 */

#define PE_SIZEOF_SHORT_NAME				 8

struct PE_SECTION_HEADER {
	byte name[PE_SIZEOF_SHORT_NAME] __attribute__ ((packed));
	dword data_vsize __attribute__ ((packed));
	dword data_address __attribute__ ((packed));
	dword data_size __attribute__	((packed));
	dword data_offset __attribute__ ((packed));
	dword relocation_offset __attribute__ ((packed));
	dword linenumber_offset __attribute__ ((packed));
	word relocation_count __attribute__ ((packed));
	word linenumber_count __attribute__ ((packed));
	dword characteristics __attribute__ ((packed));
};

#define PE_SIZEOF_SECTION_HEADER		   40

/*
 * Section characteristics.
 */

//      PE_SCN_TYPE_REG                   0x00000000  // Reserved.
//      PE_SCN_TYPE_DSECT                 0x00000001  // Reserved.
//      PE_SCN_TYPE_NOLOAD                0x00000002  // Reserved.
//      PE_SCN_TYPE_GROUP                 0x00000004  // Reserved.
#define PE_SCN_TYPE_NO_PAD			   0x00000008	// Reserved.
//      PE_SCN_TYPE_COPY                  0x00000010  // Reserved.

#define PE_SCN_CNT_CODE				   0x00000020	// Section contains code.
#define PE_SCN_CNT_INITIALIZED_DATA	   0x00000040	// Section contains initialized data.
#define PE_SCN_CNT_UNINITIALIZED_DATA	   0x00000080	// Section contains uninitialized data.

#define PE_SCN_LNK_OTHER			   0x00000100	// Reserved.
#define PE_SCN_LNK_INFO				   0x00000200	// Section contains comments or some other type of information.
//      PE_SCN_TYPE_OVER                  0x00000400  // Reserved.
#define PE_SCN_LNK_REMOVE			   0x00000800	// Section contents will not become part of image.
#define PE_SCN_LNK_COMDAT			   0x00001000	// Section contents comdat.
//                                           0x00002000  // Reserved.

//      PE_SCN_MEM_PROTECTED - Obsolete   0x00004000
#define PE_SCN_MEM_FARDATA			   0x00008000
//      PE_SCN_MEM_SYSHEAP  - Obsolete    0x00010000
#define PE_SCN_MEM_PURGEABLE			   0x00020000
#define PE_SCN_MEM_16BIT			   0x00020000
#define PE_SCN_MEM_LOCKED			   0x00040000
#define PE_SCN_MEM_PRELOAD			   0x00080000

#define PE_SCN_ALIGN_1byteS			   0x00100000	//
#define PE_SCN_ALIGN_2byteS			   0x00200000	//
#define PE_SCN_ALIGN_4byteS			   0x00300000	//
#define PE_SCN_ALIGN_8byteS			   0x00400000	//
#define PE_SCN_ALIGN_16byteS			   0x00500000	// Default alignment if no others are specified.
#define PE_SCN_ALIGN_32byteS			   0x00600000	//
#define PE_SCN_ALIGN_64byteS			   0x00700000	//
// Unused                                    0x00800000

#define PE_SCN_LNK_NRELOC_OVFL		0x01000000	// Section contains extended relocations.
#define PE_SCN_MEM_DISCARDABLE		0x02000000	// Section can be discarded.
#define PE_SCN_MEM_NOT_CACHED			0x04000000	// Section is not cachable.
#define PE_SCN_MEM_NOT_PAGED			0x08000000	// Section is not pageable.
#define PE_SCN_MEM_SHARED			0x10000000	// Section is shareable.
#define PE_SCN_MEM_EXECUTE			0x20000000	// Section is executable.
#define PE_SCN_MEM_READ				0x40000000	// Section is readable.
#define PE_SCN_MEM_WRITE				0x80000000	// Section is writeable.

/*
 *	Export
 */

struct PE_EXPORT_DIRECTORY {
	dword characteristics __attribute__ ((packed));
	time_t timestamp __attribute__ ((packed));
	word major_version __attribute__ ((packed));
	word minor_version __attribute__ ((packed));
	dword name_address __attribute__ ((packed));
	dword ordinal_base __attribute__ ((packed));
	dword function_count __attribute__ ((packed));
	dword name_count __attribute__ ((packed));
	dword function_table_address __attribute__ ((packed));
	dword name_table_address __attribute__ ((packed));
	dword ordinal_table_address __attribute__ ((packed));
};

/*
 *	Import
 */

struct PE_THUNK_DATA	{
	union {
		dword forwarder_string __attribute__ ((packed));
		dword function_desc_address __attribute__ ((packed));
		dword ordinal __attribute__ ((packed));
		dword data_address __attribute__ ((packed));
	};
};

struct PE_IMPORT_DESCRIPTOR {
	union {
		dword characteristics __attribute__ ((packed)); 		// 0 for terminating null import descriptor
		dword original_first_thunk __attribute__ ((packed));  	// rva to original unbound IAT
	};
	time_t timestamp __attribute__ ((packed));		// 0 if not bound,
											// -1 if bound, and real date\time stamp
											//     in IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT (new BIND)
											// O.W. date/time stamp of DLL bound to (Old BIND)

	dword forwarder_chain __attribute__ ((packed));  	// -1 if no forwarders
	dword name;
	dword first_thunk __attribute__ ((packed)); 		// rva to IAT (if bound this IAT has actual addresses)
};

/*
 *	Delay Import
 */

struct PE_DELAY_IMPORT_DESCRIPTOR {
	dword attributes __attribute__ ((packed));
	dword name __attribute__ ((packed));
	dword module_handle __attribute__ ((packed));
	dword delay_iat __attribute__ ((packed));
	dword delay_int __attribute__ ((packed));
	dword bound_delay_import_table __attribute__ ((packed));
	dword unload_delay_import_table __attribute__ ((packed));
	time_t timestamp __attribute__ ((packed));
};

/*
 *	Resource
 */

struct PE_RESOURCE_DIRECTORY {
	dword characteristics __attribute__ ((packed));
	time_t timedate_stamp __attribute__ ((packed));
	word major_version __attribute__ ((packed));
	word minor_version __attribute__ ((packed));
	word name_count __attribute__ ((packed));
	word id_count __attribute__ ((packed));
//    PE_RESOURCE_DIRECTORY_ENTRY DirectoryEntries[];
};

#define PE_RESOURCE_NAME_IS_STRING	   0x80000000
#define PE_RESOURCE_DATA_IS_DIRECTORY	   0x80000000

struct PE_RESOURCE_DIRECTORY_ENTRY {
	dword   name;				// or id
	dword   offset_to_directory;	// or data
};

/*typedef struct PE_RESOURCE_DIRECTORY_STRING {
    word    Length;
    CHAR    NameString[ 1 ];
};


typedef struct PE_RESOURCE_DIR_STRING_U {
    word    Length;
    WCHAR   NameString[ 1 ];
};*/

struct PE_RESOURCE_DATA_ENTRY {
    dword	  offset_to_data __attribute__ ((packed));
    dword	  size __attribute__ ((packed));
    dword	  codepage __attribute__	((packed));
    dword	  reserved __attribute__	((packed));
};

#endif /* !__PESTRUCT_H_ */
