/* 
 *	The HTE Editor
 *	x86dis.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "x86dis.h"

#include <stdio.h>
#include <string.h>

#define mkmod(modrm) (((modrm)>>6)&3)
#define mkreg(modrm) (((modrm)>>3)&7)
#define mkrm(modrm) ((modrm)&7)

#define mkscale(modrm) (((modrm)>>6)&3)
#define mkindex(modrm) (((modrm)>>3)&7)
#define mkbase(modrm) ((modrm)&7)

int modrm16_1[8] = { X86REG_BX, X86REG_BX, X86REG_BP, X86REG_BP,
				 X86REG_SI, X86REG_DI, X86REG_BP, X86REG_BX};
int modrm16_2[8] = { X86REG_SI, X86REG_DI, X86REG_SI, X86REG_DI,
				 X86REG_NO, X86REG_NO, X86REG_NO, X86REG_NO};

int sibbase[8] = { X86REG_AX, X86REG_CX, X86REG_DX, X86REG_BX,
			    X86REG_SP, X86REG_BP, X86REG_SI, X86REG_DI };

int sibindex[8] = { X86REG_AX, X86REG_CX, X86REG_DX, X86REG_BX,
				X86REG_NO, X86REG_BP, X86REG_SI, X86REG_DI };

int sibscale[4] = {1, 2, 4, 8};

/*
 *	CLASS x86dis
 */

x86dis::x86dis(int _opsize, int _addrsize)
{
	opsize=_opsize;
	addrsize=_addrsize;
}

x86dis::~x86dis()
{
}

dis_insn *x86dis::decode(byte *code, byte _maxlen, CPU_ADDR _addr)
{
	ocodep=code;
/* initialize */
	codep=ocodep;
	maxlen=_maxlen;
	addr=_addr.addr32;
	modrm=-1;
	sib=-1;
	invalid=0;
	memset(&insn, 0, sizeof(insn));
	insn.eopsize=opsize;
	insn.eaddrsize=opsize;

	prefixes();

	insn.opcode=c;
	decode_insn(&x86_insns[insn.opcode]);

	if (invalid) {
		insn.name="<invalid>";
		insn.size=1;
		for (int i=0; i<3; i++) insn.op[i].type=X86_OPTYPE_EMPTY;
	} else {
		insn.size=codep-ocodep;
	}
	return &insn;
}

void x86dis::decode_modrm(x86_insn_op *op, char size, int allow_reg, int allow_mem, int mmx)
{
	int modrm=getmodrm();
	int mod=mkmod(modrm);
	int rm=mkrm(modrm);
	if (mod==3) {
/* reg */
		if (!allow_reg) {
			invalidate();
			return;
		}
		if (mmx) {
			op->type=X86_OPTYPE_MMX;
			op->size=8;
			op->mmx=rm;
		} else {
			op->type=X86_OPTYPE_REG;
			op->size=esizeop(size);
			op->reg=rm;
		}
	} else {
/* mem */
		if (!allow_mem) {
			invalidate();
			return;
		}
		op->mem.addrsize=insn.eaddrsize;
		if (insn.eaddrsize==X86_ADDRSIZE16) {
			op->type=X86_OPTYPE_MEM;
			op->size=esizeop(size);
			op->mem.floatptr=isfloat(size);
			if ((mod==0) && (rm==6)) {
				op->mem.disp=getword();
				op->mem.base=X86REG_NO;
				op->mem.index=X86REG_NO;
				op->mem.scale=0;
			} else {
				op->mem.base=modrm16_1[rm];
				op->mem.index=modrm16_2[rm];
				op->mem.scale=1;
				switch (mod) {
					case 0:
						op->mem.disp=0;
						break;
					case 1:
						op->mem.disp=(char)getbyte();
						break;
					case 2:
						op->mem.disp=(short)getword();
						break;
				}
			}
		} else {
			op->type=X86_OPTYPE_MEM;
			op->size=esizeop(size);
			op->mem.floatptr=isfloat(size);
			if ((mod==0) && (rm==5)) {
				op->mem.disp=getdword();
				op->mem.base=X86REG_NO;
				op->mem.index=X86REG_NO;
				op->mem.scale=0;
			} else if (rm==4) {
				decode_sib(op, mod);
			} else {
				op->mem.base=rm;
				op->mem.index=X86REG_NO;
				op->mem.scale=1;
				switch (mod) {
					case 0:
						op->mem.disp=0;
						break;
					case 1:
						op->mem.disp=(char)getbyte();
						break;
					case 2:
						op->mem.disp=getdword();
						break;
				}
			}
		}
	}
}

void x86dis::decode_insn(x86opc_insn *xinsn)
{
	if (!xinsn->name) {
		x86opc_insn_op_special special=*((x86opc_insn_op_special*)(&xinsn->op[0]));
		switch (special.type) {
			case SPECIAL_TYPE_INVALID:
				invalidate();
				break;
			case SPECIAL_TYPE_PREFIX:
				switch (c) {
					case 0x0f:
						insn.opcode=getbyte();
						insn.opcodeclass=X86DIS_OPCODE_CLASS_EXT;
						decode_insn(&x86_insns_ext[insn.opcode]);
						break;
					default:
						invalidate();
						break;
				}
				break;
			case SPECIAL_TYPE_GROUP: {
				int m=mkreg(getmodrm());
				insn.opcode|=m<<8;
				decode_insn(&x86_group_insns[special.data][m]);
				break;
			}
			case SPECIAL_TYPE_FGROUP: {
				int m=getmodrm();
				if (mkmod(m)==3) {
					x86opc_finsn f=x86_float_group_insns[special.data][mkreg(m)];
					if (f.single) {
						decode_insn(&f.insn);
					} else if (f.group) {
						decode_insn(&f.group[mkrm(m)]);
					} else invalidate();
				} else {
					decode_insn(&x86_modfloat_group_insns[special.data][mkreg(m)]);
				}
				break;
			}
		}
	} else {
		insn.name=xinsn->name;
		for (int i=0; i<3; i++) {
			decode_op(&insn.op[i], &xinsn->op[i]);
		}
	}
}

void x86dis::decode_op(x86_insn_op *op, x86opc_insn_op *xop)
{
	switch (xop->type) {
		case TYPE_0:
			return;
		case TYPE_A: {
/* direct address without ModR/M */
			op->type=X86_OPTYPE_IMM;
			op->size=esizeaddr(xop->size);
			op->mem.floatptr=isfloat(xop->size);
			switch (op->size) {
				case 2:
					op->mem.disp=getword();
					break;
				case 4:
					op->mem.disp=getdword();
					break;
			}
			break;
		}
		case TYPE_C: {
/* reg of ModR/M picks control register */
			op->type=X86_OPTYPE_CRX;
			op->size=esizeop(xop->size);
			op->crx=mkreg(getmodrm());
			break;
		}
		case TYPE_D: {
/* reg of ModR/M picks debug register */
			op->type=X86_OPTYPE_DRX;
			op->size=esizeop(xop->size);
			op->drx=mkreg(getmodrm());
			break;
		}
		case TYPE_E: {
/* ModR/M (general reg or memory) */
			decode_modrm(op, xop->size, 1, 1, 0);
			break;
		}
		case TYPE_F: {
/* r/m of ModR/M picks a fpu register */
			op->type=X86_OPTYPE_STX;
			op->size=10;
			op->stx=mkrm(getmodrm());
			break;
		}
		case TYPE_Fx: {
/* extra picks a fpu register */
			op->type=X86_OPTYPE_STX;
			op->size=10;
			op->stx=xop->extra;
			break;
		}
		case TYPE_G: {
/* reg of ModR/M picks general register */
			op->type=X86_OPTYPE_REG;
			op->size=esizeop(xop->size);
			op->reg=mkreg(getmodrm());
			break;
		}
		case TYPE_Is: {
/* signed immediate */
			op->type=X86_OPTYPE_IMM;
			op->size=esizeop(xop->extendedsize);
			int s=esizeop(xop->size);
			switch (s) {
				case 1:
					op->imm=(char)getbyte();
					break;
				case 2:
					op->imm=(short)getword();
					break;
				case 4:
					op->imm=getdword();
					break;
			}
			switch (op->size) {
				case 1:
					op->imm|=0xff;
					break;
				case 2:
					op->imm|=0xffff;
					break;
			}
			break;
		}
		case TYPE_I: {
/* unsigned immediate */
			op->type=X86_OPTYPE_IMM;
			op->size=esizeop(xop->extendedsize);
			int s=esizeop(xop->size);
			switch (s) {
				case 1:
					op->imm=getbyte();
					break;
				case 2:
					op->imm=getword();
					break;
				case 4:
					op->imm=getdword();
					break;
			}
			break;
		}
		case TYPE_Ix: {
/* fixed immediate */
			op->type=X86_OPTYPE_IMM;
			op->size=esizeop(xop->extendedsize);
			op->imm=xop->extra;
			break;
		}
		case TYPE_J: {
/* relative branch offset */
			op->type=X86_OPTYPE_IMM;
			op->size=esizeaddr(xop->extendedsize);
			int s=esizeop(xop->size);
			switch (s) {
				case 1:
					op->imm=(char)getbyte()+addr;
					break;
				case 2:
					op->imm=(short)getword()+addr;
					break;
				case 4:
					op->imm=getdword()+addr;
					break;
			}
			break;
		}
		case TYPE_M: {
/* ModR/M (memory only) */
			decode_modrm(op, xop->size, 0, 1, 0);
			break;
		}
		case TYPE_O: {
/* direct memory without ModR/M */
			op->type=X86_OPTYPE_MEM;
			op->size=esizeop(xop->size);
			op->mem.floatptr=isfloat(xop->size);
			op->mem.addrsize=insn.eaddrsize;
			switch (insn.eaddrsize) {
				case X86_ADDRSIZE16:
					op->mem.disp=getword();
					break;
				case X86_ADDRSIZE32:
					op->mem.disp=getdword();
					break;
			}
			op->mem.base=X86REG_NO;
			op->mem.index=X86REG_NO;
			op->mem.scale=1;
			break;
		}
		case TYPE_P: {
/* reg of ModR/M picks MMX register */
			op->type=X86_OPTYPE_MMX;
			op->size=8;
			op->mmx=mkreg(getmodrm());
			break;
		}
		case TYPE_Q: {
/* ModR/M (MMX reg or memory) */
			decode_modrm(op, xop->size, 1, 1, 1);
			break;
		}
		case TYPE_R: {
/* rm of ModR/M picks general register */
			op->type=X86_OPTYPE_REG;
			op->size=esizeop(xop->size);
			op->reg=mkrm(getmodrm());
			break;
		}
		case TYPE_Rx: {
/* extra picks register */
			op->type=X86_OPTYPE_REG;
			op->size=esizeop(xop->size);
			op->reg=xop->extra;
			break;
		}
		case TYPE_S: {
/* reg of ModR/M picks segment register */
			op->type=X86_OPTYPE_SEG;
			op->size=esizeop(xop->size);
			op->seg=mkreg(getmodrm());
			break;
		}
		case TYPE_Sx: {
/* extra picks segment register */
			op->type=X86_OPTYPE_SEG;
			op->size=esizeop(xop->size);
			op->seg=xop->extra;
			break;
		}
		case TYPE_T: {
/* reg of ModR/M picks test register */
			op->type=X86_OPTYPE_TRX;
			op->size=esizeop(xop->size);
			op->trx=mkreg(getmodrm());
			break;
		}
	}
}

void x86dis::decode_sib(x86_insn_op *op, int mod)
{
	int sib=getsib();
	int scale=mkscale(sib);
	int index=mkindex(sib);
	int base=mkbase(sib);
	int disp=mod;
	if (base==5) {
		if (mod==0) {
			op->mem.base=X86REG_NO;
			disp=2;
		} else {
			op->mem.base=X86REG_BP;
		}
		op->mem.index=sibindex[index];
		op->mem.scale=sibscale[scale];
	} else {
		op->mem.base=sibbase[base];
		op->mem.index=sibindex[index];
		op->mem.scale=sibscale[scale];
	}
	switch (disp) {
		case 0:
			op->mem.disp=0;
			break;
		case 1:
			op->mem.disp=(char)getbyte();
			break;
		case 2:
			op->mem.disp=getdword();
			break;
	}
}

int x86dis::esizeaddr(char c)
{
	switch (c) {
		case SIZE_B:
			return 1;
		case SIZE_W:
			return 2;
		case SIZE_D:
			return 4;
		case SIZE_Q:
			return 8;
		case SIZE_S:
			return 4;
		case SIZE_L:
			return 8;
		case SIZE_T:
			return 10;
		case SIZE_C:
			if (insn.eaddrsize==X86_ADDRSIZE16) return 1; else return 2;
		case SIZE_V:
			if (insn.eaddrsize==X86_ADDRSIZE16) return 2; else return 4;
		case SIZE_P:
			if (insn.eaddrsize==X86_ADDRSIZE16) return 4; else return 6;
	}
	return 0;
}

int x86dis::esizeop(char c)
{
	switch (c) {
		case SIZE_B:
			return 1;
		case SIZE_W:
			return 2;
		case SIZE_D:
			return 4;
		case SIZE_Q:
			return 8;
		case SIZE_S:
			return 4;
		case SIZE_L:
			return 8;
		case SIZE_T:
			return 10;
		case SIZE_C:
			if (insn.eopsize==X86_OPSIZE16) return 1; else return 2;
		case SIZE_V:
			if (insn.eopsize==X86_OPSIZE16) return 2; else return 4;
		case SIZE_P:
			if (insn.eopsize==X86_OPSIZE16) return 4; else return 6;
	}
	return 0;
}

byte x86dis::getbyte()
{
	if (codep-ocodep+1<=maxlen) {
		addr+=1;
		return *(codep++);
	} else {
		invalidate();
		return 0;
	}
}

word x86dis::getword()
{
	if (codep-ocodep+2<=maxlen) {
		addr+=2;
		return *(((word*)codep)++);
	} else {
		invalidate();
		return 0;
	}
}

dword x86dis::getdword()
{
	if (codep-ocodep+4<=maxlen) {
		addr+=4;
		return *(((dword*)codep)++);
	} else {
		invalidate();
		return 0;
	}
}

int x86dis::getmodrm()
{
	if (modrm==-1) modrm=getbyte();
	return modrm;
}

int x86dis::getsib()
{
	if (sib==-1) sib=getbyte();
	return sib;
}

byte x86dis::getsize(dis_insn *disasm_insn)
{
	return ((x86dis_insn*)disasm_insn)->size;
}

void x86dis::hexd(char **s, int size, int style, int imm)
{
	char ff[16];
	char *f=(char*)&ff;
	char *t=*s;
	f+=sprintf(f, "%%");
	if (style & X86DIS_STYLE_HEX_CSTYLE) f+=sprintf(f, "#");
	if (!(style & X86DIS_STYLE_HEX_NOZEROPAD)) f+=sprintf(f, "0%d", size);
	if (style & X86DIS_STYLE_HEX_UPPERCASE) f+=sprintf(f, "X"); else
		 f+=sprintf(f, "x");
	if (style & X86DIS_STYLE_HEX_ASMSTYLE) f+=sprintf(f, "h");
	(*s)+=sprintf(*s, ff, imm);
	if ((style & X86DIS_STYLE_HEX_NOZEROPAD) && (*t-'0'>9)) {
		memmove(t+1, t, strlen(t)+1);
		*t='0';
		(*s)++;
	}
}

void x86dis::invalidate()
{
	invalid=1;
}

int x86dis::isfloat(char c)
{
	switch (c) {
		case SIZE_S:
		case SIZE_L:
		case SIZE_T:
			return 1;
	}
	return 0;
}

void x86dis::prefixes()
{
	bool notprefix=0;
	insn.lockprefix=X86DIS_PREFIX_NO;
	insn.repprefix=X86DIS_PREFIX_NO;
	insn.segprefix=X86DIS_PREFIX_NO;
	while ((codep-ocodep<15) && (!notprefix)) {
		c=getbyte();
		switch (c) {
			case 0x26:
				insn.segprefix=X86DIS_PREFIX_ES;
				break;
			case 0x2e:
				insn.segprefix=X86DIS_PREFIX_CS;
				break;
			case 0x36:
				insn.segprefix=X86DIS_PREFIX_SS;
				break;
			case 0x3e:
				insn.segprefix=X86DIS_PREFIX_DS;
				break;
			case 0x64:
				insn.segprefix=X86DIS_PREFIX_FS;
				break;
			case 0x65:
				insn.segprefix=X86DIS_PREFIX_GS;
				break;
			case 0x66:
				insn.eopsize=(opsize==X86_OPSIZE16) ? X86_OPSIZE32 : X86_OPSIZE16;
				break;
			case 0x67:
				insn.eaddrsize=(addrsize==X86_ADDRSIZE16) ? X86_ADDRSIZE32 : X86_ADDRSIZE16;
				break;
			case 0xf0:
				insn.lockprefix=X86DIS_PREFIX_LOCK;
				break;
			case 0xf2:
				insn.repprefix=X86DIS_PREFIX_REPZ;
				break;
			case 0xf3:
				insn.repprefix=X86DIS_PREFIX_REPNZ;
				break;
			default:
				notprefix=1;
		}
	}
}

void x86dis::str_op(char *opstr, x86dis_insn *insn, x86_insn_op *op, int style)
{
	switch (op->type) {
		case X86_OPTYPE_IMM: {
			CPU_ADDR a;
			a.addr32=op->imm;
			char *s=addr_sym_func(a);
			if (s) {
				strcpy(opstr, s);
			} else {
				char *g=opstr;
				switch (op->size) {
					case 1:
						hexd(&g, 2, style, op->imm);
						break;
					case 2:
						hexd(&g, 4, style, op->imm);
						break;
					case 4:
						hexd(&g, 8, style, op->imm);
						break;
				}
			}
			break;
		}
		case X86_OPTYPE_REG: {
			int j;
			switch (op->size) {
				case 1: j=0; break;
				case 2: j=1; break;
				case 4: j=2; break;
				default:
					j=-1;
			}
			if (j!=-1) sprintf(opstr, x86_regs[j][op->reg]);
			break;
		}
		case X86_OPTYPE_SEG:
			sprintf(opstr, x86_segs[op->seg]);
			break;
		case X86_OPTYPE_CRX:
			sprintf(opstr, "cr%d", op->crx);
			break;
		case X86_OPTYPE_DRX:
			sprintf(opstr, "dr%d", op->drx);
			break;
		case X86_OPTYPE_TRX:
			sprintf(opstr, "tr%d", op->trx);
			break;
		case X86_OPTYPE_STX:
			if (op->stx) {
				sprintf(opstr, "st(%d)", op->stx);
			} else {
				sprintf(opstr, "st");
			}
			break;
		case X86_OPTYPE_MMX:
			sprintf(opstr, "mm%d", op->mmx);
			break;
		case X86_OPTYPE_MEM: {
			char *d=opstr;
			int reg;
			if (style & X86DIS_STYLE_EXPLICIT_MEMSIZE) {
				if (op->mem.floatptr) {
					switch (op->size) {
						case 4:
							d+=sprintf(d, "single ptr ");
							break;
						case 8:
							d+=sprintf(d, "double ptr ");
							break;
						case 10:
							d+=sprintf(d, "extended ptr ");
							break;
					}
				} else {
					switch (op->size) {
						case 1:
							d+=sprintf(d, "byte ptr ");
							break;
						case 2:
							d+=sprintf(d, "word ptr ");
							break;
						case 4:
							d+=sprintf(d, "dword ptr ");
							break;
						case 6:
							d+=sprintf(d, "fword ptr ");
							break;
						case 8:
							d+=sprintf(d, "qword ptr ");
							break;
					}
				}
			}
			if (insn->eaddrsize==X86_ADDRSIZE16) reg=1; else reg=2;
			if (insn->segprefix!=X86DIS_PREFIX_NO) d+=sprintf(d, "%s:", x86_segs[insn->segprefix]);
			d+=sprintf(d, "[");

			int optimize_addr=style & X86DIS_STYLE_OPTIMIZE_ADDR;
			int first=1;
			if (optimize_addr && (op->mem.base!=X86REG_NO) && (op->mem.base==op->mem.index)) {
				d+=sprintf(d, "%s*%d", x86_regs[reg][op->mem.index], op->mem.scale+1);
				first=0;
			} else {
				if (op->mem.base!=X86REG_NO) {
					d+=sprintf(d, "%s", x86_regs[reg][op->mem.base]);
					first=0;
				}
				if (op->mem.index!=X86REG_NO) {
					if (!first) d+=sprintf(d, "+");
					if (optimize_addr && (op->mem.scale==1)) {
						d+=sprintf(d, "%s", x86_regs[reg][op->mem.index]);
					} else {
						d+=sprintf(d, "%s*%d", x86_regs[reg][op->mem.index], op->mem.scale);
					}
					first=0;
				}
			}
			if (!optimize_addr || (optimize_addr && op->mem.disp) || first) {
				CPU_ADDR a;
				a.addr32=op->mem.disp;
				char *s=addr_sym_func(a);
				if (s) {
					if (!first) d+=sprintf(d, "+");
					d+=sprintf(d, "%s", s);
				} else {
					switch (op->mem.addrsize) {
						case X86_ADDRSIZE16:
							if (!first) {
								if (op->mem.disp&0x80000000) {
									d+=sprintf(d, "-");
									op->mem.disp=-op->mem.disp;
								} else d+=sprintf(d, "+");
							}
							hexd(&d, 4, style, op->mem.disp);
							break;
						case X86_ADDRSIZE32:
							if (!first) {
								if (op->mem.disp&0x80000000) {
									d+=sprintf(d, "-");
									op->mem.disp=-op->mem.disp;
								} else d+=sprintf(d, "+");
							}
							hexd(&d, 8, style, op->mem.disp);
							break;
					}
				}
			}
			d+=sprintf(d, "]");
			break;
		}
		default:
			opstr[0]=0;
	}
}

void x86dis::str_format(char **str, char **format, char *p, char *n, char *op[3], char stopchar, int print)
{
	char *f=*format;
	char *s=*str;
	while (*f) {
		if (*f==stopchar) break;
		switch (*f) {
			case '\t':
				if (print) do *(s++)=' '; while ((s-insnstr) % X86DIS_STYLE_TABSIZE);
				break;
			case DISASM_STRF_VAR:
				f++;
				if (print) {
					char *t=0;
					switch (*f) {
						case DISASM_STRF_PREFIX:
							t=p;
							break;
						case DISASM_STRF_NAME:
							t=n;
							break;
						case DISASM_STRF_FIRST:
							t=op[0];
							break;
						case DISASM_STRF_SECOND:
							t=op[1];
							break;
						case DISASM_STRF_THIRD:
							t=op[2];
							break;
					}
					s+=sprintf(s, t);
				}
				break;
			case DISASM_STRF_COND: {
				char *t=0;
				f++;
				switch (*f) {
					case DISASM_STRF_PREFIX:
						t=p;
						break;
					case DISASM_STRF_NAME:
						t=n;
						break;
					case DISASM_STRF_FIRST:
						t=op[0];
						break;
					case DISASM_STRF_SECOND:
						t=op[1];
						break;
					case DISASM_STRF_THIRD:
						t=op[2];
						break;
				}
				f+=2;
				if ((t) && (t[0])) {
					str_format(&s, &f, p, n, op, *(f-1), 1);
				} else {
					str_format(&s, &f, p, n, op, *(f-1), 0);
				}
				break;
			}
			default:
				if (print) *(s++)=*f;
		}
		f++;
	}
	*s=0;
	*format=f;
	*str=s;
}

char *x86dis::str(dis_insn *disasm_insn, int style)
{
	return strf(disasm_insn, style, DISASM_STRF_DEFAULT_FORMAT);
}

char *x86dis::strf(dis_insn *disasm_insn, int style, char *format)
{
	x86dis_insn *insn=(x86dis_insn*)disasm_insn;
	char prefix[64];
	char *p=prefix;
	*p=0;
	if (insn->lockprefix==X86DIS_PREFIX_LOCK) p+=sprintf(p, "lock ");
	if (insn->repprefix==X86DIS_PREFIX_REPZ) {
		p+=sprintf(p, "repz ");
	} else if (insn->repprefix==X86DIS_PREFIX_REPNZ) {
		p+=sprintf(p, "repnz ");
	}
	if ((p!=prefix) && (*(p-1)==' ')) {
		p--;
		*p=0;
	}

	char ops[3][512];
	char *op[3];
	for (int i=0; i<3; i++) {
		op[i]=(char*)&ops[i];
		str_op(op[i], insn, &insn->op[i], style);
	}
	char *s=insnstr;
	str_format(&s, &format, prefix, insn->name, op, 0, 1);
	return insnstr;
}

